import os
import json

path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)

#  global environment
with open(dir_path + os.sep + "environment.json", 'r') as f:
    environment_global = json.load(f)

#  local environment
environment_local = dict()
if os.path.exists(dir_path + os.sep + "environment_local.json"):
    with open(dir_path + os.sep + "environment_local.json", 'r') as f:
        environment_local = json.load(f)

# update dicts
environment = {**environment_global, **environment_local}


class Enviroment:
    path = __file__
    path = path.replace("environment.py", "")

    artap_root = path
    tests_root = os.path.abspath(path + "." + os.sep + "tests" + os.sep)
            
    condor_host = environment["condor_host"]
    condor_login = environment["condor_login"]
                
    ssh_host = environment["ssh_host"]
    ssh_login = environment["ssh_login"]

    comsol_path = environment["comsol_path"]

    loopback_ip = environment["loopback_ip"]
    server_initial_port = environment["server_initial_port"]
    server_keep_live_delay = environment["server_keep_live_delay"]



import unittest

from artap.problem import Problem
from artap.algorithm_bayesopt import BayesOptSerial, BayesOptParallel
from artap.algorithm_bayes import Bayes, BayesianOptimization

from artap.results import Results
from artap.benchmark_functions import Booth

import numpy as np


# PBOUNDS = {'p1': (-10, 10), 'p2': (-10, 10)}
PBOUNDS = {'x': (-2, 10)}


def target_func(x):
    # val = sum(x)
    # val = -Booth.eval(x)
    val = np.exp(-(x - 2)**2) + np.exp(-(x - 6)**2/10) + 1/ (x**2 + 1)
    # print(val, x)
    return val


class MyProblemOne(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "TestBayesOptParallel"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-2, 10]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector
        return -(np.exp(-(x - 2)**2) + np.exp(-(x - 6)**2/10) + 1 / (x**2 + 1))


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "TestBayesOptSerial"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
                      {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        return [Booth.eval(individual.vector)]


class TestBayesOptOptimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def xtest_local_problem_bayesopt_parallel(self):
        problem = MyProblem()
        algorithm = BayesOptParallel(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 100
        algorithm.run()
        # TODO - multiprocess test

        results = Results(problem)
        optimum = results.find_minimum(name='F')
        self.assertAlmostEqual(optimum, 0, places=2)

    def xtest_local_problem_bayesopt_serial(self):
        problem = MyProblem()
        algorithm = Bayes(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 200
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum(name='F')
        self.assertAlmostEqual(optimum, 0, places=2)


    def xtest_local_problem_bayesopt_serial(self):
        problem = MyProblem()
        algorithm = BayesOptSerial(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 20
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum(name='F')
        print(optimum)


    def xtest_local_problem_bayesop_one(self):
        problem = MyProblemOne()
        algorithm = BayesOptSerial(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 20
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum(name='F')
        print(optimum)

    def test_bayesoptimization(self):
        optimizer = BayesianOptimization(f=target_func,
                                         pbounds=PBOUNDS,
                                         random_state=np.random.RandomState(1))

        optimizer.maximize(init_points=5, n_iter=30, acq="ucb", kappa=10.0)
        print("maximum")
        print(optimizer.max)


if __name__ == '__main__':
    unittest.main()



from .individual import Individual


class Population:
    def __init__(self, individuals: list = None, archives: list = None):
        if individuals is None:
            self.individuals = []
        else:
            self.individuals = individuals.copy()

        if archives is None:
            self.archives = []
        else:
            self.archives = archives.copy()

        self.pareto_vectors: list = []
        self.pareto_costs: list = []

    def __str__(self):
        string = "individuals: {}\n".format(len(self.individuals))
        for individual in self.individuals:
            string += "\t{},\n".format(individual.__str__())
        return string

    def gen_population_from_table(self, table):
        for parameters in table:
            individual = Individual(parameters)
            self.individuals.append(individual)

    def to_list(self, population_number):
        table = []
        for individual in self.individuals:
            params = individual.to_list()
            params.insert(0, self.individuals.index(individual) + 1)
            # population index
            params.insert(1, population_number)
            table.append(params)

        return table



import sqlite3
import os
import json
import time
import logging

from abc import abstractmethod

from .individual import Individual
from .population import Population


class SqliteHandler(logging.Handler):
    """
    Thread-safe logging handler for SQLite.
    """

    def __init__(self, data_store):
        logging.Handler.__init__(self)

        self.data_store = data_store
        self.data_store.create_structure_log()

    def emit(self, record: logging.LogRecord):
        """

        self.format(record)
        # format record
        record.dbtime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))
        if record.exc_info:  # for exceptions
            record.exc_text = logging._defaultFormatter.formatException(record.exc_info)
        else:
            record.exc_text = ""
        """

        # Insert the log record
        try:
            connection = sqlite3.connect(self.data_store.database_name)

            cursor = connection.cursor()
            cursor.execute("INSERT INTO log(timestamp, source, loglevel, loglevelname, message, args, module, funcname, lineno, exception, process, threadname) "
                           "VALUES (:timestamp, :source, :loglevel, :loglevelname, :message, :args, :module, :funcname, :lineno, :exception, :process, :threadname)",
                           { "timestamp": str(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))),
                             "source": str(record.name),
                             "loglevel": record.levelno,
                             "loglevelname": str(record.levelname),
                             "message": str(record.getMessage()),
                             "args": str(record.args),
                             "module": str(record.module),
                             "funcname": str(record.funcName),
                             "lineno": record.lineno,
                             "exception": str(record.exc_text),
                             "process": record.process,
                             "threadname": str(record.threadName) })
            connection.commit()
            cursor.close()

            connection.close()
        except sqlite3.Error as e:
            print("SqliteHandler: error occurred:", e.args[0])


class DataStore:
    """ Class  ensures saving data from optimization problems. """

    def __init__(self, problem):
        self.problem = problem

        # create the new loop and worker thread
        # self.worker_loop = asyncio.new_event_loop()
        # worker = Thread(target=self._start_worker, args=(self.worker_loop,))
        # worker.daemon = True
        # # Start the thread
        # worker.start()

        # populations
        self.populations = []

    def __del__(self):
        # print("DataStore:def __del__(self):")
        pass

    # def _start_worker(self, loop):
    #     asyncio.set_event_loop(loop)
    #     loop.run_forever()

    @abstractmethod
    def create_structure(self):
        pass

    @abstractmethod
    def write_individual(self, individual):
        # TODO: check this hack
        if len(self.populations) == 0:
            self.populations.append(Population())

        # add individual
        self.populations[-1].individuals.append(individual)

    @abstractmethod
    def write_population(self, population):
        self.populations.append(population)

    def get_id(self):
        return 0

"""
class SqliteDataStore(DataStore):
    def __init__(self, problem, database_name=None, remove_existing=True):
        super().__init__(problem)

        self.database_name = database_name

        # remove database and create structure
        if remove_existing and os.path.exists(self.database_name):
            os.remove(self.database_name)

        if remove_existing:
            self.create_structure()
        else:
            self.read_from_datastore()

        # set datastore to problem
        self.problem.data_store = self

        # set db log
        if self.problem.options['log_db_handler']:
            # create file handler and set level to debug
            file_handler = SqliteHandler(self)
            file_handler.setLevel(logging.DEBUG)
            # add formatter to SqliteHandler
            file_handler.setFormatter(self.problem.formatter)
            # add SqliteHandler to logger
            self.problem.logger.addHandler(file_handler)

    def __del__(self):
        super().__del__()

    def _execute_command(self, command):
        """  """
        self._execute_command_async(self.database_name, command)
        # self.worker_loop.call_soon_threadsafe(self._execute_command_async, self.database_name, command)

    @staticmethod
    def _execute_command_async(database_name, command):
        # print(database_name)
        # print(command)

        try:
            connection = sqlite3.connect(database_name)
            # connection.execute('pragma journal_mode=wal')

            cursor = connection.cursor()
            cursor.execute(command)
            connection.commit()
            cursor.close()

            connection.close()
        except sqlite3.Error as e:
            print("SqliteDataStore: execute: error occurred:", e.args[0])

    def _create_structure_task(self):
        exec_cmd = 'CREATE TABLE IF NOT EXISTS problem (INTEGER PRIMARY KEY, ' \
                   'name TEXT, ' \
                   'description TEXT,' \
                   'algorithm TEXT,' \
                   'calculation_time NUMBER,' \
                   'state TEXT)'
        self._execute_command(exec_cmd)

        exec_cmd = "INSERT INTO problem(name, description, state) " \
                   "values('%s', '%s', '%s');" % (self.problem.name, self.problem.description, "running")
        self._execute_command(exec_cmd)

    def _create_structure_individual(self):
        exec_cmd = 'CREATE TABLE IF NOT EXISTS data (id INTEGER, population_id INTEGER, '

        for parameter in self.problem.parameters.keys():
            exec_cmd += parameter + " NUMBER, "

        for cost in self.problem.costs:
            exec_cmd += cost + " NUMBER,"

        exec_cmd += 'front_number NUMBER, crowding_distance NUMBER, feasible NUMBER, depends_on NUMBER, ' \
                    'modified_param NUMBER, dominates JSON, gradient JSON '
        exec_cmd += ");"

        self._execute_command(exec_cmd)

    def _create_structure_parameters(self):
        exec_cmd = 'CREATE TABLE IF NOT EXISTS parameters (id INTEGER PRIMARY KEY AUTOINCREMENT, name TEXT,' \
                   'initial_value NUMBER, low_boundary NUMBER, high_boundary NUMBER, precision NUMBER);'
        self._execute_command(exec_cmd)

        for parameter in self.problem.parameters.items():
            exec_cmd = "INSERT INTO parameters(name, initial_value, low_boundary, high_boundary, precision) " \
                       "VALUES( '{}', {}, {}, {}, {});".format(parameter[0], parameter[1]['initial_value'],
                                                               parameter[1]['bounds'][0], parameter[1]['bounds'][1],
                                                               parameter[1]['precision'] if 'precision' in parameter[1] else 0)
            self._execute_command(exec_cmd)

    def _create_structure_costs(self):
        exec_cmd = 'CREATE TABLE IF NOT EXISTS costs (id INTEGER PRIMARY KEY AUTOINCREMENT, name TEXT);'
        self._execute_command(exec_cmd)

        for cost in self.problem.costs:
            exec_cmd = "INSERT INTO costs(name) values('%s');" % cost
            self._execute_command(exec_cmd)

    def create_structure(self):
        self._create_structure_task()
        self._create_structure_individual()
        self._create_structure_parameters()
        self._create_structure_costs()

    def create_structure_log(self):
        exec_cmd = 'CREATE TABLE IF NOT EXISTS log(' \
                    'timestamp TEXT,' \
                    'source TEXT,' \
                    'loglevel INT,' \
                    'loglevelname TEXT,' \
                    'message TEXT,' \
                    'args TEXT,' \
                    'module TEXT,' \
                    'funcname TEXT,' \
                    'lineno INT,' \
                    'exception TEXT,' \
                    'process INT,' \
                    'threadname TEXT)'

        self._execute_command(exec_cmd)

    def write_individual(self, individual):
        # add individual
        super().write_individual(individual)

        # add to database
        if self.problem.options['save_level'] == "individual":
            params = individual.to_list()
            # individual index
            params.insert(0, self.populations[-1].individuals.index(individual) + 1)
            # population index
            params.insert(1, len(self.populations))

            exec_cmd = "INSERT INTO data VALUES ("

            for i in range(len(params) - 1):
                if params[i] == float('inf'):
                    params[i] = -1
                if type(params[i]) == list:
                    par = "'" + str(params[i]) + "'"
                else:
                    par = str(params[i])
                exec_cmd += " " + par + ","

            if type(params[i]) == list:
                par = "'" + str(params[i]) + "'"
            else:
                par = str(params[i])
            exec_cmd += " " + par + ")"

            self._execute_command(exec_cmd)

    def write_population(self, population):
        super().write_population(population)

        # write to database
        connection = sqlite3.connect(self.database_name)
        table = population.to_list(self.populations.index(population))
        cursor = connection.cursor()
        for params in table:
            exec_cmd = "INSERT INTO data VALUES ("
            for i in range(len(params) - 1):
                if params[i] == float('inf'):
                    params[i] = -1
                if type(params[i]) == list:
                    par = "'" + str(params[i]) + "'"
                else:
                    par = str(params[i])
                exec_cmd += " " + par + ","

            if type(params[i]) == list:
                par = "'" + str(params[i]) + "'"
            else:
                par = str(params[i])
            exec_cmd += " " + par + ")"

            cursor.execute(exec_cmd)
        cursor.close()
        connection.commit()
        connection.close()

    def read_from_datastore(self):
        connection = sqlite3.connect(self.database_name)
        # connection.execute('pragma journal_mode=wal')

        cursor = connection.cursor()

        exec_cmd_problem = "SELECT * FROM problem"
        problem_table = cursor.execute(exec_cmd_problem).fetchall()
        self.problem.name = problem_table[0][1]

        # parameters
        exec_cmd_params = "SELECT * FROM parameters"
        columns = []
        for row in cursor.execute(exec_cmd_params):
            columns.append(row)

        self.problem.parameters = {}
        for parameter in columns:
            self.problem.parameters[parameter[1]] = {'initial_value': parameter[2],
                                                     'bounds': [parameter[3], parameter[4]],
                                                     'precision': parameter[5]}
        exec_cmd_costs = "SELECT * FROM costs"

        # costs
        costs = []
        for cost in cursor.execute(exec_cmd_costs):
            costs.append(cost[1])
        self.problem.costs = {cost: 0.0 for cost in costs}

        exec_cmd_data = "SELECT * FROM data"
        data = cursor.execute(exec_cmd_data)

        # clear populations and individuals
        self.individuals = []
        self.populations = []
        # new population
        population = Population()
        self.populations.append(population)

        table = list()
        for row in data:
            table.append(row)

        cursor.close()
        connection.close()

        table.sort(key=lambda x: x[1])  # Sorting according to population number

        current_population = table[0][1]
        for row in table:
            if row[1] == current_population:
                individual = Individual(list(row[2:2 + len(self.problem.parameters)]))
                l = 2 + len(self.problem.parameters) + len(self.problem.costs)
                individual.costs = row[2 + len(self.problem.parameters): 2 + len(self.problem.parameters) + len(self.problem.costs)]
                individual.front_number = row[l]
                # individual.crowding_distance = row[l+1]
                # individual.feasible = row[l+2]
                individual.depends_on = row[l+3]
                individual.modified_param = row[l+4]
                individual.dominate = json.loads(row[l+5])
                individual.gradient = json.loads(row[l+6])

                # individual.dominate = json.loads(row[l+3])
                # individual.gradient = json.loads(row[l+4])


                # add to population
                population.individuals.append(individual)
            else:
                # increase population
                self.populations.append(population)
                population = Population()
                current_population = row[1]
"""

class FileDataStore(DataStore):

    def __init__(self, problem, database_name=None, remove_existing=True, mode="write", backend="shelve"):
        super().__init__(problem)

        self.database_name = database_name

        if remove_existing and mode == "write":
            if os.path.exists(self.database_name):
                os.remove(self.database_name)

        if backend == "shelve":
            import shelve

            if mode == "write":
                self.db = shelve.open(self.database_name, flag='c', writeback=True)
                # remove database and create structure
                if remove_existing and os.path.exists(self.database_name):
                    self.create_structure()
            else:
                self.db = shelve.open(self.database_name, flag='r')
                self.read_from_datastore()
        elif backend == "diskcache":
            import diskcache

            self.db = diskcache.Cache(directory=self.database_name)
            if mode == "write":
                # remove database and create structure
                if remove_existing and os.path.exists(self.database_name):
                    self.create_structure()
            else:
                self.read_from_datastore()
        elif backend == "sqlitedict":
            from .environment import Enviroment
            import sys
            sys.path.append(Enviroment.artap_root + os.sep + "../3rdparty" + os.sep + "sqlitedict")
            from sqlitedict import SqliteDict

            self.db = SqliteDict(self.database_name, autocommit=True)
            if mode == "write":
                # remove database and create structure
                if remove_existing and os.path.exists(self.database_name):
                    self.create_structure()
            else:
                self.read_from_datastore()
        else:
            assert 1

        # set datastore to problem
        self.problem.data_store = self

    def __del__(self):
        super().__del__()

    def create_structure(self):
        self.db["name"] = self.problem.name
        self.db["description"] = self.problem.description
        self.db["parameters"] = self.problem.parameters
        self.db["costs"] = self.problem.costs

        self.db["populations"] = []

    def write_individual(self, individual):
        # add individual
        super().write_individual(individual)

        # TODO: check this hack
        populations = self.db["populations"]
        if len(populations) == 0:
            populations.append(Population())
            self.db["populations"] = populations

        # add individual
        populations = self.db["populations"]
        populations[-1].individuals.append(individual)
        self.db["populations"] = populations
        # self.db["populations"][-1].append(individual)
        # self.db.sync()

    def write_population(self, population):
        super().write_population(population)

        # write to database
        populations = self.db["populations"]
        populations.append(Population())
        self.db["populations"] = populations
        # self.db["populations"].append(population)
        # self.db.sync()

    def read_from_datastore(self):
        self.problem.name = self.db["name"]
        self.problem.description = self.db["description"]
        self.problem.parameters = self.db["parameters"]
        self.problem.costs = self.db["costs"]

        self.populations = self.db["populations"]


class DummyDataStore(DataStore):

    def __init__(self, problem=None):
        super().__init__(problem)
        self.create_structure()

    def __del__(self):
        super().__del__()

    def create_structure(self):
        pass

    def write_individual(self, individual):
        super().write_individual(individual)

    def write_population(self, population):
        super().write_population(population)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 14 09:50:41 2019

@author: david
"""

import scipy.io
import os
import numpy as np
from scipy.sparse import linalg as la
import pylab as pl
from artap.dynamical_systems import Model, DiscreteLTISystem


class POD:
    def __init__(self, stifness_matrix, mass_matrix, rhs_vector, solution):
        self.stifness_matrix = stifness_matrix
        self.mass_matrix = mass_matrix
        self.rhs = rhs_vector
        self.solution = solution
        self.steps = 80
        self.num = 3
        self.total_time = 30
        self.dt = self.total_time / self.steps

    def reduce(self, index):
        mass_matrix = self.mass_matrix / self.dt
        matrix_lhs = mass_matrix + self.stifness_matrix
        slnt = np.zeros([len(rhs), self.steps])
        for i in range(1, self.steps):
            slnt[:, i] = la.spsolve(matrix_lhs, mass_matrix @ slnt[:, i-1] + rhs[:, 0]*signal(i * self.dt))

        pl.plot(slnt[index, :])
        pl.show()
        c = np.dot(slnt, slnt.T)
        [D, V] = (np.linalg.eigh(c))

        L = np.zeros([len(V), self.num])
        L[:, :] = V[:, -self.num:]

        mor_rhs = np.dot(L.T, self.rhs)
        mor_mass = L.T @ mass_matrix @ L

        mor_lhs = L.T @ matrix_lhs @ L
        matrix_a = np.linalg.inv(mor_lhs) @ mor_mass
        matrix_b = np.linalg.inv(mor_lhs) @ mor_rhs
        matrix_c = L[index, :]
        return matrix_a, matrix_b, matrix_c

    @staticmethod
    def find_index(solution, px, py):
        ind = -1
        val = 1e6
        for i in range(len(solution)):
            tmp = (px - solution[i, 0])**2 + (py - solution[i, 1])**2
            if tmp < val:
                val = tmp
                ind = i

        print(ind, val)
        return ind


def signal(t):
    u = 1
    if t <= 7.5:
        u = 1.7 * 2
    if 7.5 < t <= 10:
        u = 0.8 * 1.7 * 2
    if t > 10:
        u = 0.1 * 1.7 * 2
    return [u]


# ToDo: Make tests from this
if __name__ == "__main__":
    working_dir = "/home/david/Dokumenty/Git/Software/artap-projects/brazing_mor"
    current_path = os.getcwd()
    os.chdir(working_dir)
    stiffness = scipy.io.loadmat('transient_solver-heat_matrix_stiffness.mat')['matrix_stiffness']
    mass = scipy.io.loadmat('transient_solver-heat_matrix_mass.mat')['matrix_mass']
    rhs = scipy.io.loadmat('transient_solver-heat_rhs.mat')['rhs']
    other = scipy.io.loadmat('transient_solver-heat_other.mat')['other']

    pod = POD(stiffness, mass, rhs, other)
    A, B, C = pod.reduce(10)
    print(A)
    print(B)
    print(C)
    D = [0]

    x0 = [0, 0, 0]
    plant = Model(dt=30/80)
    system = DiscreteLTISystem(plant, A, B, C, D, x0, dt=30/80)
    system.connect(signal)
    plant.run(30)
    system.plot('output')
    pl.show()


import unittest

from artap.problem import Problem
from artap.algorithm_sensitivity import Sensitivity


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "LocalPythonProblem"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5], 'precision': 1e-1},
                           {'name': 'x_2', 'initial_value': 2.5, 'bounds': [2.2, 2.4], 'precision': 1e-1},
                           {'name': 'x_3', 'initial_value': 2.5, 'bounds': [0, 5], 'precision': 1e-1}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector
        result = x[0] * x[0] + 0.1 * x[1] * x[1] + x[2] * x[2]
        return [result]


class TestSensitivity(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem(self):
        problem = MyProblem()
        algorithm = Sensitivity(problem, ['x_2', 'x_3'])
        algorithm.options['max_population_size'] = 10
        algorithm.run()


if __name__ == '__main__':
    unittest.main()



from artap.problem import Problem
from artap.results import Results
from artap.algorithm_bayesopt import BayesOptSerial
from artap.algorithm_nlopt import NLopt, LN_BOBYQA, LN_COBYLA
from artap.algorithm_genetic import NSGAII

from agrossuite import agros as a2d


class Capacitor(Problem):
    def set(self):
        self.name = "AgrosProblem"
        self.parameters = [{'name': 'r1', 'initial_value': 0.02, 'bounds': [0.01, 0.03]},
                      {'name': 'r2', 'initial_value': 0.04, 'bounds': [0.035, 0.1]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        # problem
        problem = a2d.problem(clear=True)
        problem.coordinate_type = "planar"
        problem.mesh_type = "triangle"

        # project parameters
        problem.parameters["r1"] = individual.vector[0]
        problem.parameters["r2"] = individual.vector[1]
        problem.parameters["eps"] = 2.5

        # fields
        # electrostatic
        electrostatic = problem.field("electrostatic")
        electrostatic.analysis_type = "steadystate"
        electrostatic.matrix_solver = "umfpack"
        electrostatic.number_of_refinements = 1
        electrostatic.polynomial_order = 1
        electrostatic.adaptivity_type = "disabled"
        electrostatic.solver = "linear"

        # boundaries
        electrostatic.add_boundary("GND", "electrostatic_potential", {"electrostatic_potential": 0})
        electrostatic.add_boundary("V", "electrostatic_potential", {"electrostatic_potential": 1})
        electrostatic.add_boundary("neumann", "electrostatic_surface_charge_density",
                                   {"electrostatic_surface_charge_density": 0})

        # materials
        electrostatic.add_material("dielectric",
                                   {"electrostatic_permittivity": "eps", "electrostatic_charge_density": 0})

        # geometry
        geometry = problem.geometry()
        geometry.add_edge("r1", 0, "r2", 0, boundaries={"electrostatic": "neumann"})
        geometry.add_edge("r2", 0, 0, "r2", angle=90, boundaries={"electrostatic": "GND"})
        geometry.add_edge(0, "r2", 0, "r1", boundaries={"electrostatic": "neumann"})
        geometry.add_edge("r1", 0, 0, "r1", angle=90, boundaries={"electrostatic": "V"})

        geometry.add_label("(r1 + r2) / 2.", "(r2 - r1) / 2", materials={"electrostatic": "dielectric"})

        computation = problem.computation()
        computation.solve()

        solution = computation.solution("electrostatic")
        result = solution.volume_integrals()["We"]

        capacitance_req = 80.  # pF
        capacitance = 4.0 * 2.0 * result * 1e12  # pF
        print('Evaluated capacitance: {}, required capacitance: {}, difference: {}'.format(capacitance, capacitance_req, abs(capacitance - capacitance_req)))

        return [abs(capacitance - capacitance_req)]


def bobyqa():
    problem = Capacitor()
    algorithm = NLopt(problem)
    algorithm.options['n_iterations'] = 30
    algorithm.options['algorithm'] = LN_BOBYQA
    algorithm.options['verbose_level'] = 0
    algorithm.run()

    results = Results(problem)
    optimum = results.find_minimum('F')
    print(optimum)


def bayesopt():
    problem = Capacitor()
    algorithm = BayesOptSerial(problem)
    algorithm.options['n_iterations'] = 10
    algorithm.options['verbose_level'] = 0
    algorithm.run()

    results = Results(problem)
    optimum = results.find_minimum('F')
    print(optimum)


def nsga2():
    problem = Capacitor()
    algorithm = NSGAII(problem)
    algorithm.options['max_population_number'] = 10
    algorithm.options['max_population_size'] = 10
    algorithm.run()

    results = Results(problem)
    optimum = results.find_minimum('F')
    print(optimum)

# nsga2()
bobyqa()
# bayesopt()


from artap.problem import Problem
from artap.algorithm_genetic import NSGAII
from artap.results import Results

import optproblems as optp
import optproblems.dtlz as dtlz

import matplotlib.pyplot as plt
#
# TODO::: No module named : diversipy
#

class DTLZ2_TEST_Problems(Problem):
    """
    This example shows the solution of a DTLZ test example. This name is an abbreviation of the
    authors Deb -- Thiele -- Laumanns -- Zitzler (I -- VII). The definition and the solution of
    these multiobjective optimization test problems defined in [Deb2002] and in the optproblems
    library.

    Here, the DTLZ -- 2 problem was solved for 10 independent parameters. The solution, the Pareto-front
    of this example is placed on a 3 dimensional sphere.
    NSGA - II algorithm is used to solve this example.

    References
    ----------

    .. [Deb2002] Deb, K.; Thiele, L.; Laumanns, M.; Zitzler, E. (2002).
        Scalable multi-objective optimization test problems, Proceedings of
        the IEEE Congress on Evolutionary Computation, pp. 825-830

    """


    def set(self):

        self.name = "DTLZ -- 2 test problem"

        self.parameters = [{'name':'x_1', 'bounds': [0, 1]},
                           {'name':'x_2', 'bounds': [0, 1]},
                           {'name':'x_3', 'bounds': [0, 1]},
                           {'name':'x_4', 'bounds': [0, 1]},
                           {'name': 'x_5', 'bounds': [0, 1]},
                           {'name': 'x_6', 'bounds': [0, 1]},
                           {'name': 'x_7', 'bounds': [0, 1]},
                           {'name': 'x_8', 'bounds': [0, 1]},
                           {'name': 'x_9', 'bounds': [0, 1]},
                           {'name': 'x_10', 'bounds': [0, 1]}]

        self.costs = [{'name': 'F_1'},{'name': 'F_2'}]

    def evaluate(self, x):
        # objective values were stored together with decision variables
        problem = optp.Problem(dtlz.DTLZ2(3, 10), num_objectives=3)
        solutions = [optp.Individual(x)]

        problem.batch_evaluate(solutions)
        return solutions[0].objective_values.copy()


def test_problem_2():

    problem = DTLZ2_TEST_Problems()
    algorithm = NSGAII(problem)
    algorithm.options['max_population_number'] = 1000
    algorithm.options['max_population_size'] = 100
    algorithm.run()

    b = Results(problem)
    solution = b.pareto_values()

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.scatter([s[0] for s in solution],
                [s[1] for s in solution],
                [s[2] for s in solution])
    ax.set_xlim([0, 1.1])
    ax.set_ylim([0, 1.1])
    ax.set_zlim([0, 1.1])

    plt.xlabel("$f_1(x)$")
    plt.ylabel("$f_2(x)$")

    ax.view_init(elev=30.0, azim=15.0)
    plt.show()

def test_problem_1():

    problem = DTLZ2_TEST_Problems("DTLZ2")
    algorithm = NSGAII(problem)
    algorithm.options['max_population_number'] = 100
    algorithm.options['max_population_size'] = 100
    algorithm.run()

    b = Results(problem)

    solution = b.pareto_values()

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.scatter([s[0] for s in solution],
                [s[1] for s in solution],
                [s[2] for s in solution])
    ax.set_xlim([0, 1.1])
    ax.set_ylim([0, 1.1])
    ax.set_zlim([0, 1.1])

    plt.xlabel("$f_1(x)$")
    plt.ylabel("$f_2(x)$")

    ax.view_init(elev=30.0, azim=15.0)
    plt.show()


test_problem_2()


"""
 Module is dedicated to describe optimization problem.
"""

from .datastore import DataStore, DummyDataStore
from .utils import ConfigDictionary
from .surrogate import SurrogateModelEval
from abc import abstractmethod

import logging
import datetime
import tempfile
import os
import shutil

from enum import Enum


# ToDo: Inhere executors, remove enum

class ProblemType(Enum):
    comsol = 0
    analytical = 1
    agros = 2
    matlab = 3
    python = 4

CRITICAL = logging.CRITICAL
FATAL = logging.FATAL
ERROR = logging.ERROR
WARNING = logging.WARNING
WARN = logging.WARN
INFO = logging.INFO
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET

_log_level = [CRITICAL, ERROR, WARNING, INFO, DEBUG]


class Problem:
    """ The Class Problem Is a main class which collects information about optimization task """

    __is_frozen = False

    def __init__(self):

        self.options = ConfigDictionary()
        # options
        self.options.declare(name='calculate_gradients', default=False,
                             desc='calculate gradient for individuals')
        self.options.declare(name='save_level', default="population",
                             desc='Save level')
        self.options.declare(name='log_level', default=logging.INFO, values=_log_level,
                             desc='Log level')
        self.options.declare(name='log_file_handler', default=False,
                             desc='Enable file handler')
        self.options.declare(name='log_db_handler', default=True,
                             desc='Enable DB handler')
        self.options.declare(name='log_console_handler', default=True,
                             desc='Enable console handler')
        self.options.declare(name='time_out', default=600,
                             desc='Maximal time for calculation')
        self.options.declare(name='save_data_files', default=False,
                             desc='Saving data from computation')

        self.name: str = str()
        self.description: str = str()
        self.parameters: dict = dict()
        self.costs: dict = dict()
        self.data_store: DataStore
        self.type = "None"
        self.working_dir = tempfile.mkdtemp() + os.sep
        self.output_files = None
        self.executor = None
        self.data_store = DummyDataStore(self)
        self.is_working_dir_set = True

        # tmp name
        d = datetime.datetime.now()
        ts = d.strftime("{}-%f".format(self.__class__.__name__))

        # create logger
        self.logger = logging.getLogger(ts)
        self.logger.setLevel(self.options['log_level'])

        for h in list(self.logger.handlers):
            print(h)

        # create formatter
        self.formatter = logging.Formatter('%(asctime)s (%(levelname)s): '
                                           '%(name)s - %(funcName)s (%(lineno)d) - %(message)s')

        if self.options['log_console_handler']:
            # create console handler and set level to debug
            stream_handler = logging.StreamHandler()
            stream_handler.setLevel(logging.DEBUG)
            # add formatter to StreamHandler
            stream_handler.setFormatter(self.formatter)
            # add StreamHandler to logger
            self.logger.addHandler(stream_handler)

        # self.id = self.data_store.get_id()

        self.logger.debug("START")
        # self.logger.debug('This message should go to the log file')
        # self.logger.info('So should this')
        # self.logger.warning('And this, too')

        # surrogate model (default - only simple eval)
        self.surrogate = SurrogateModelEval(self)
        self.signs = []
        self._freeze()
        self.set()
        for cost in self.costs:
            if 'criteria' in cost:
                if cost['criteria'] == 'minimize':
                    self.signs.append(1)
                else:
                    self.signs.append(-1)
            else:
                self.signs.append(1)

    def __del__(self):
        if not self.is_working_dir_set:
            shutil.rmtree(self.working_dir)

    @abstractmethod
    def set(self):
        pass

    def parameters_len(self):
        return len(self.parameters)

    # search for and defines the initial values for the optimization problem
    def get_initial_values(self):
        values = []
        for parameter in self.parameters:
            if 'initial_value' in parameter:
                values.append(parameter['initial_value'])
            else:
                values.append(0)
        return values

    def get_parameter_types(self):
        # TODO: is it important? or a typecheck is enough?
        p_types = []
        for parameter in self.parameters:

            if 'parameter_type' in parameter:
                if parameter['parameter_type'].lower == 'real':
                    p_types.append('real')

                if parameter['parameter_type'].lower == 'integer':
                    p_types.append('integer')

                if parameter['parameter_type'].lower == 'boolean':
                    p_types.append('boolean')

            else:
                p_types.append('real')

        return p_types

    @abstractmethod
    def evaluate(self, x: list):
        """ :param x: list of the variables """
        pass

    def evaluate_constraints(self, x: list):
        """ :param x: list of the variables """
        pass

    def __setattr__(self, key, value):
        if self.__is_frozen and not hasattr(self, key):
            raise TypeError(" %r is a frozen class" % self)
        object.__setattr__(self, key, value)

        # working dir must be set
        if hasattr(self.options, 'log_file_handler'):
            if self.working_dir:
                # create file handler and set level to debug
                file_handler = logging.FileHandler(self.working_dir + "/data.log")
                file_handler.setLevel(logging.DEBUG)
                # add formatter to FileHandler
                file_handler.setFormatter(self.formatter)
                # add FileHandler to logger
                self.logger.addHandler(file_handler)
                self.is_working_dir_set = True

    def _freeze(self):
        self.__is_frozen = True



import os
import unittest
from artap.problem import Problem
from artap.datastore import FileDataStore
from artap.algorithm_nlopt import NLopt
from artap.algorithm_nlopt import LN_BOBYQA

from artap.results import Results
from artap.benchmark_functions import Booth

import tempfile


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "NLopt_BOBYQA"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        return [Booth.eval(individual.vector)]


class TestDataStoreFile(unittest.TestCase):

    def test_read_dbm_data_store(self):
        problem = Problem()
        results = Results(problem)
        database_name = "." + os.sep + "data" + os.sep + "data.db"
        problem.data_store = FileDataStore(problem, database_name=database_name, mode="read")
        optimum = results.find_minimum('F')
        self.assertAlmostEqual(optimum.costs[0], 1.854, 3)
        self.assertEqual(problem.name, 'NLopt_BOBYQA')
        self.assertEqual(problem.description, '')
        self.assertEqual(problem.parameters[0]['initial_value'], 2.5)
        self.assertEqual(problem.parameters[0]['bounds'][0], -10)
        self.assertEqual(problem.parameters[0]['bounds'][1], 10)
        self.assertEqual(problem.parameters[1]['initial_value'], 1.5)
        self.assertEqual(problem.parameters[1]['bounds'][0], -10)
        self.assertEqual(problem.parameters[1]['bounds'][1], 10)
        individual = problem.data_store.populations[0].individuals[6]
        self.assertAlmostEqual(individual.vector[0], 2.6989845414318134, 5)
        self.assertAlmostEqual(individual.vector[1], 1.83185007711266, 5)
        self.assertAlmostEqual(individual.costs[0], 5.378264283347013, 5)

    def test_write_dbm_data_store(self):
        problem = MyProblem()

        # set data store
        database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".db").name
        problem.data_store = FileDataStore(problem, database_name=database_name, mode="write")
        algorithm = NLopt(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['n_iterations'] = 10
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F')
        self.assertAlmostEqual(optimum.costs[0], 1.854, 3)

        problem.data_store.db.close()

        # check db
        import shelve

        db = shelve.open(problem.data_store.database_name, flag='r')
        populations = db["populations"]

        self.assertAlmostEqual(populations[-1].individuals[9].costs[0], 1.854, 3) # result
        db.close()

        # remove file
        # print(database_name)
        os.remove(database_name)

    def test_write_sqlitedict_data_store(self):
        problem = MyProblem()

        # set data store
        database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".sqlite").name
        problem.data_store = FileDataStore(problem, database_name=database_name, backend="sqlitedict", mode="write")
        algorithm = NLopt(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['n_iterations'] = 10
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F')
        self.assertAlmostEqual(optimum.costs[0], 1.854, 3)

        problem.data_store.db.close()

        # check db
        from artap.environment import Enviroment
        import sys
        sys.path.append(Enviroment.artap_root + os.sep + "../3rdparty" + os.sep + "sqlitedict")
        from sqlitedict import SqliteDict

        db = SqliteDict(database_name, autocommit=True)

        populations = db["populations"]

        self.assertAlmostEqual(populations[-1].individuals[9].costs[0], 1.854, 3) # result
        db.close()

        # remove file
        # print(database_name)
        os.remove(database_name)


if __name__ == '__main__':
    unittest.main()



from .problem import Problem
from .algorithm import Algorithm
from .population import Population
from .individual import GeneticIndividual
from .operators import RandomGeneration, SimulatedBinaryCrossover, \
    PmMutation, TournamentSelection, GradientGeneration, ParetoDominance, EpsilonDominance

import time
import sys

from copy import deepcopy

EPSILON = sys.float_info.epsilon


class GeneralEvolutionaryAlgorithm(Algorithm):
    """ Basis Class for evolutionary algorithms """

    def __init__(self, problem: Problem, name="General Evolutionary Algorithm"):
        super().__init__(problem, name)
        self.problem = problem

        self.generator = None
        self.selector = None
        self.mutator = None
        self.crossover = None

    def run(self):
        pass


class GeneticAlgorithm(GeneralEvolutionaryAlgorithm):

    def __init__(self, problem: Problem, name="General Evolutionary Algorithm"):
        super().__init__(problem, name)
        self.parameters_length = len(self.problem.parameters)

        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum evaluations')
        self.options.declare(name='max_population_number', default=10, lower=1,
                             desc='max_population_number')
        self.options.declare(name='max_population_size', default=100, lower=1,
                             desc='Maximal number of individuals in population')

    def generate(self, parents, archive=None):
        """ generate two children from two different parents """

        # Preserve bounds of Pareto-Front
        children = []
        # removed_parents = []
        # for i in range(len(self.problem.signs)):
        #     bound = min(parents, key=lambda x: (self.problem.signs[i] * x.costs[i]))
        #     children.append(deepcopy(bound))
        #     parents.remove(bound)
        #     removed_parents.append(deepcopy(bound))
        while len(children) < self.population_size:
            parent1 = self.selector.select(parents)

            if archive is not None:
                parent2 = self.selector.select(archive)
            else:
                parent2 = self.selector.select(parents)

            while parent1 == parent2:
                parent2 = self.selector.select(parents)

            # crossover
            child1, child2 = self.crossover.cross(parent1, parent2)

            # mutation
            child1 = self.mutator.mutate(child1)
            child2 = self.mutator.mutate(child2)
            if not any(child1 == item for item in children):
                children.append(deepcopy(child1))  # Always create new individual
            if not any(child1 == item for item in children):
                children.append(deepcopy(child2))  # Always create new individual

        # parents.extend(removed_parents)
        return children

    def run(self):
        pass


class NSGAII(GeneticAlgorithm):

    def __init__(self, problem: Problem, name="NSGA_II Evolutionary Algorithm"):
        super().__init__(problem, name)

        self.problem.options["save_level"] = "population"

        self.options.declare(name='prob_cross', default=0.6, lower=0,
                             desc='prob_cross')
        self.options.declare(name='prob_mutation', default=0.2, lower=0,
                             desc='prob_mutation')

        # TODO: Make new class NSGAII with gradients
        self.gradient_generator = None

    def run(self):
        # set random generator
        self.generator = RandomGeneration(self.problem, individual_class=GeneticIndividual)
        self.generator.init(self.options['max_population_size'])

        if self.options['calculate_gradients'] is True:
            self.gradient_generator = GradientGeneration(self.problem.parameters)

        # set crossover
        # self.crossover = SimpleCrossover(self.problem.parameters, self.options['prob_cross'])
        self.crossover = SimulatedBinaryCrossover(self.problem.parameters, self.options['prob_cross'])

        # set mutator
        # self.mutator = SimpleMutation(self.problem.parameters, self.options['prob_mutation'])
        self.mutator = PmMutation(self.problem.parameters, self.options['prob_mutation'])

        # set selector
        self.selector = TournamentSelection(self.problem.parameters, self.problem.signs)

        # create initial population and evaluate individuals
        population = self.gen_initial_population()
        # non-dominated sort of individuals
        self.selector.sorting(population.individuals)
        self.selector.crowding_distance(population.individuals)
        # write to data store

        self.problem.data_store.write_population(population)

        if self.options['calculate_gradients'] is True:
            self.gradient_generator.init(population.individuals)
            gradient_population = Population(self.gradient_generator.generate())
            self.evaluate(gradient_population.individuals)
            self.evaluate_gradient(population.individuals, gradient_population.individuals)
            self.problem.data_store.write_population(population)

        t_s = time.time()
        self.problem.logger.info(
            "NSGA_II: {}/{}".format(self.options['max_population_number'], self.options['max_population_number'] * self.population_size))

        # optimization
        for it in range(self.options['max_population_number']):
            # generate new offsprings
            offsprings = self.generate(population.individuals)
            # evaluate the offsprings
            self.evaluate(offsprings)

            # if (self.options['calculate_gradients'] is True) and population.number > 20:
            #     population.evaluate_gradients()

            # add the parents to the offsprings
            offsprings.extend(deepcopy(population.individuals))

            # non-dominated truncate on the guys
            self.selector.sorting(offsprings)
            self.selector.crowding_distance(offsprings)

            # sort offsprings and ### remove duplicates by set!!!
            parents = sorted(set(offsprings), key=lambda x: (x.front_number, -x.crowding_distance))

            # truncate
            offsprings = parents[:self.population_size]

            # write population
            population = Population(offsprings)
            self.problem.data_store.write_population(population)

            if self.options['calculate_gradients'] is True:
                self.gradient_generator.init(population.individuals)
                gradient_population = Population(self.gradient_generator.generate())
                self.evaluate(gradient_population.individuals)
                self.evaluate_gradient(population.individuals, gradient_population.individuals)
                self.problem.data_store.write_population(population)

        t = time.time() - t_s
        self.problem.logger.info("NSGA_II: elapsed time: {} s".format(t))


class EpsMOEA(GeneticAlgorithm):

    def __init__(self, problem: Problem, name="EpsMOEA Algorithm"):
        super().__init__(problem, name)

        self.options.declare(name='prob_cross', default=0.6, lower=0,
                             desc='prob_cross')
        self.options.declare(name='prob_mutation', default=0.2, lower=0,
                             desc='prob_mutation'),
        self.options.declare(name='epsilons', default=0.01, lower=1e-6,
                             desc='prob_epsilons')

    def run(self):
        # set random generator
        self.generator = RandomGeneration(self.problem, individual_class=GeneticIndividual)  # the same as in the case of NSGA-II
        self.generator.init(self.options['max_population_size'])

        # set crossover
        self.crossover = SimulatedBinaryCrossover(self.problem.parameters, self.options['prob_cross'])
        self.mutator = PmMutation(self.problem.parameters, self.options['prob_mutation'])

        # Part A: non-dominated sort of individuals
        # -----
        Selector_Pareto = TournamentSelection(self.problem.parameters, self.problem.signs)
        self.selector = TournamentSelection(self.problem.parameters, self.problem.signs)
        # the same as in the case of NSGA - ii
        # this operator is used to generate the new individuals

        # create initial population and evaluate individuals
        population = self.gen_initial_population(True)  # archiving True

        Selector_Pareto.sorting(population.individuals)
        Selector_Pareto.crowding_distance(population.individuals)

        # Part B: eps-dominated sort of the individuals with archiving
        # -----
        Selector_EpsDom = TournamentSelection(self.problem.parameters, self.problem.signs,
                                              dominance=EpsilonDominance, epsilons=self.options['epsilons'])
        Selector_EpsDom.sorting(population.archives)
        Selector_EpsDom.crowding_distance(population.archives)

        # write to data store
        self.problem.data_store.write_population(population) # TODO: modify the database connection to handle the archives

        t_s = time.time()
        self.problem.logger.info(
            "Eps-MOEA: {}/{}".format(self.options['max_population_number'], self.population_size))

        # optimization
        for it in range(self.options['max_population_number']):

            # generate and evaluate the next generation
            child = self.generate(population.individuals, population.archives)
            self.evaluate(child)

            arch_child = deepcopy(child)
            # PART A
            # non-dominated sorting of the newly generated and the older guys like in NSGA-ii
            # add the parents to the offsprings
            child.extend(population.individuals)

            # non-dominated truncate on the guys
            self.selector.sorting(child)
            Selector_Pareto.crowding_distance(child)

            parents = sorted(set(child), key=lambda x: (x.front_number, -x.crowding_distance))
            child = parents[:self.population_size]  # truncate

            # eps dominated truncate on the guys
            Selector_EpsDom.sorting(arch_child)
            Selector_EpsDom.crowding_distance(arch_child)

            arch_parents = sorted(set(arch_child), key=lambda x: (x.front_number, -x.crowding_distance))
            arch_child = arch_parents[:self.population_size]  # truncate

            # write population
            population = Population(child, arch_child)
            # TODO: modify the database connection to handle the archives
            self.problem.data_store.write_population(population)

        t = time.time() - t_s
        self.problem.logger.info("Eps-MOEA: {} s".format(t))



from .problem import Problem
from .algorithm import Algorithm
from .environment import Enviroment
from .job import JobSimple

import time

import sys
sys.path.append(Enviroment.artap_root + "/lib/")

import _nlopt

import builtins as __builtin__


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0


class SwigPyIterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SwigPyIterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SwigPyIterator, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _nlopt.delete_SwigPyIterator
    __del__ = lambda self: None

    def value(self):
        return _nlopt.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _nlopt.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _nlopt.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _nlopt.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _nlopt.SwigPyIterator_equal(self, x)

    def copy(self):
        return _nlopt.SwigPyIterator_copy(self)

    def next(self):
        return _nlopt.SwigPyIterator_next(self)

    def __next__(self):
        return _nlopt.SwigPyIterator___next__(self)

    def previous(self):
        return _nlopt.SwigPyIterator_previous(self)

    def advance(self, n):
        return _nlopt.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _nlopt.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _nlopt.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _nlopt.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _nlopt.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _nlopt.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _nlopt.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self
SwigPyIterator_swigregister = _nlopt.SwigPyIterator_swigregister
SwigPyIterator_swigregister(SwigPyIterator)

class nlopt_doublevector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, nlopt_doublevector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, nlopt_doublevector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _nlopt.nlopt_doublevector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _nlopt.nlopt_doublevector___nonzero__(self)

    def __bool__(self):
        return _nlopt.nlopt_doublevector___bool__(self)

    def __len__(self):
        return _nlopt.nlopt_doublevector___len__(self)

    def __getslice__(self, i, j):
        return _nlopt.nlopt_doublevector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _nlopt.nlopt_doublevector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _nlopt.nlopt_doublevector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _nlopt.nlopt_doublevector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _nlopt.nlopt_doublevector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _nlopt.nlopt_doublevector___setitem__(self, *args)

    def pop(self):
        return _nlopt.nlopt_doublevector_pop(self)

    def append(self, x):
        return _nlopt.nlopt_doublevector_append(self, x)

    def empty(self):
        return _nlopt.nlopt_doublevector_empty(self)

    def size(self):
        return _nlopt.nlopt_doublevector_size(self)

    def swap(self, v):
        return _nlopt.nlopt_doublevector_swap(self, v)

    def begin(self):
        return _nlopt.nlopt_doublevector_begin(self)

    def end(self):
        return _nlopt.nlopt_doublevector_end(self)

    def rbegin(self):
        return _nlopt.nlopt_doublevector_rbegin(self)

    def rend(self):
        return _nlopt.nlopt_doublevector_rend(self)

    def clear(self):
        return _nlopt.nlopt_doublevector_clear(self)

    def get_allocator(self):
        return _nlopt.nlopt_doublevector_get_allocator(self)

    def pop_back(self):
        return _nlopt.nlopt_doublevector_pop_back(self)

    def erase(self, *args):
        return _nlopt.nlopt_doublevector_erase(self, *args)

    def __init__(self, *args):
        this = _nlopt.new_nlopt_doublevector(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def push_back(self, x):
        return _nlopt.nlopt_doublevector_push_back(self, x)

    def front(self):
        return _nlopt.nlopt_doublevector_front(self)

    def back(self):
        return _nlopt.nlopt_doublevector_back(self)

    def assign(self, n, x):
        return _nlopt.nlopt_doublevector_assign(self, n, x)

    def resize(self, *args):
        return _nlopt.nlopt_doublevector_resize(self, *args)

    def insert(self, *args):
        return _nlopt.nlopt_doublevector_insert(self, *args)

    def reserve(self, n):
        return _nlopt.nlopt_doublevector_reserve(self, n)

    def capacity(self):
        return _nlopt.nlopt_doublevector_capacity(self)
    __swig_destroy__ = _nlopt.delete_nlopt_doublevector
    __del__ = lambda self: None
nlopt_doublevector_swigregister = _nlopt.nlopt_doublevector_swigregister
nlopt_doublevector_swigregister(nlopt_doublevector)


ForcedStop = _nlopt.ForcedStop
RoundoffLimited = _nlopt.RoundoffLimited
__version__ = str(_nlopt.version_major())+'.'+str(_nlopt.version_minor())+'.'+str(_nlopt.version_bugfix())


def nlopt_get_initial_step(opt, dx):
    return _nlopt.nlopt_get_initial_step(opt, dx)
nlopt_get_initial_step = _nlopt.nlopt_get_initial_step
GN_DIRECT = _nlopt.GN_DIRECT
GN_DIRECT_L = _nlopt.GN_DIRECT_L
GN_DIRECT_L_RAND = _nlopt.GN_DIRECT_L_RAND
GN_DIRECT_NOSCAL = _nlopt.GN_DIRECT_NOSCAL
GN_DIRECT_L_NOSCAL = _nlopt.GN_DIRECT_L_NOSCAL
GN_DIRECT_L_RAND_NOSCAL = _nlopt.GN_DIRECT_L_RAND_NOSCAL
GN_ORIG_DIRECT = _nlopt.GN_ORIG_DIRECT
GN_ORIG_DIRECT_L = _nlopt.GN_ORIG_DIRECT_L
GD_STOGO = _nlopt.GD_STOGO
GD_STOGO_RAND = _nlopt.GD_STOGO_RAND
LD_LBFGS_NOCEDAL = _nlopt.LD_LBFGS_NOCEDAL
LD_LBFGS = _nlopt.LD_LBFGS
LN_PRAXIS = _nlopt.LN_PRAXIS
LD_VAR1 = _nlopt.LD_VAR1
LD_VAR2 = _nlopt.LD_VAR2
LD_TNEWTON = _nlopt.LD_TNEWTON
LD_TNEWTON_RESTART = _nlopt.LD_TNEWTON_RESTART
LD_TNEWTON_PRECOND = _nlopt.LD_TNEWTON_PRECOND
LD_TNEWTON_PRECOND_RESTART = _nlopt.LD_TNEWTON_PRECOND_RESTART
GN_CRS2_LM = _nlopt.GN_CRS2_LM
GN_MLSL = _nlopt.GN_MLSL
GD_MLSL = _nlopt.GD_MLSL
GN_MLSL_LDS = _nlopt.GN_MLSL_LDS
GD_MLSL_LDS = _nlopt.GD_MLSL_LDS
LD_MMA = _nlopt.LD_MMA
LN_COBYLA = _nlopt.LN_COBYLA
LN_NEWUOA = _nlopt.LN_NEWUOA
LN_NEWUOA_BOUND = _nlopt.LN_NEWUOA_BOUND
LN_NELDERMEAD = _nlopt.LN_NELDERMEAD
LN_SBPLX = _nlopt.LN_SBPLX
LN_AUGLAG = _nlopt.LN_AUGLAG
LD_AUGLAG = _nlopt.LD_AUGLAG
LN_AUGLAG_EQ = _nlopt.LN_AUGLAG_EQ
LD_AUGLAG_EQ = _nlopt.LD_AUGLAG_EQ
LN_BOBYQA = _nlopt.LN_BOBYQA
GN_ISRES = _nlopt.GN_ISRES
AUGLAG = _nlopt.AUGLAG
AUGLAG_EQ = _nlopt.AUGLAG_EQ
G_MLSL = _nlopt.G_MLSL
G_MLSL_LDS = _nlopt.G_MLSL_LDS
LD_SLSQP = _nlopt.LD_SLSQP
LD_CCSAQ = _nlopt.LD_CCSAQ
GN_ESCH = _nlopt.GN_ESCH
GN_AGS = _nlopt.GN_AGS
NUM_ALGORITHMS = _nlopt.NUM_ALGORITHMS
FAILURE = _nlopt.FAILURE
INVALID_ARGS = _nlopt.INVALID_ARGS
OUT_OF_MEMORY = _nlopt.OUT_OF_MEMORY
ROUNDOFF_LIMITED = _nlopt.ROUNDOFF_LIMITED
FORCED_STOP = _nlopt.FORCED_STOP
SUCCESS = _nlopt.SUCCESS
STOPVAL_REACHED = _nlopt.STOPVAL_REACHED
FTOL_REACHED = _nlopt.FTOL_REACHED
XTOL_REACHED = _nlopt.XTOL_REACHED
MAXEVAL_REACHED = _nlopt.MAXEVAL_REACHED
MAXTIME_REACHED = _nlopt.MAXTIME_REACHED
class roundoff_limited(Exception):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, roundoff_limited, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, roundoff_limited, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _nlopt.new_roundoff_limited()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _nlopt.delete_roundoff_limited
    __del__ = lambda self: None
roundoff_limited_swigregister = _nlopt.roundoff_limited_swigregister
roundoff_limited_swigregister(roundoff_limited)

class forced_stop(Exception):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, forced_stop, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, forced_stop, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _nlopt.new_forced_stop()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _nlopt.delete_forced_stop
    __del__ = lambda self: None
forced_stop_swigregister = _nlopt.forced_stop_swigregister
forced_stop_swigregister(forced_stop)

class opt(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, opt, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, opt, name)
    __repr__ = _swig_repr
    __swig_destroy__ = _nlopt.delete_opt
    __del__ = lambda self: None

    def __init__(self, *args):
        this = _nlopt.new_opt(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def optimize(self, *args):
        return _nlopt.opt_optimize(self, *args)

    def last_optimize_result(self):
        return _nlopt.opt_last_optimize_result(self)

    def last_optimum_value(self):
        return _nlopt.opt_last_optimum_value(self)

    def get_algorithm(self):
        return _nlopt.opt_get_algorithm(self)

    def get_algorithm_name(self):
        return _nlopt.opt_get_algorithm_name(self)

    def get_dimension(self):
        return _nlopt.opt_get_dimension(self)

    def set_min_objective(self, *args):
        return _nlopt.opt_set_min_objective(self, *args)

    def set_max_objective(self, *args):
        return _nlopt.opt_set_max_objective(self, *args)

    def remove_inequality_constraints(self):
        return _nlopt.opt_remove_inequality_constraints(self)

    def remove_equality_constraints(self):
        return _nlopt.opt_remove_equality_constraints(self)

    def add_inequality_constraint(self, *args):
        return _nlopt.opt_add_inequality_constraint(self, *args)

    def add_equality_constraint(self, *args):
        return _nlopt.opt_add_equality_constraint(self, *args)

    def add_inequality_mconstraint(self, *args):
        return _nlopt.opt_add_inequality_mconstraint(self, *args)

    def add_equality_mconstraint(self, *args):
        return _nlopt.opt_add_equality_mconstraint(self, *args)

    def get_lower_bounds(self, *args):
        return _nlopt.opt_get_lower_bounds(self, *args)

    def set_lower_bounds(self, *args):
        return _nlopt.opt_set_lower_bounds(self, *args)

    def get_upper_bounds(self, *args):
        return _nlopt.opt_get_upper_bounds(self, *args)

    def set_upper_bounds(self, *args):
        return _nlopt.opt_set_upper_bounds(self, *args)

    def get_stopval(self):
        return _nlopt.opt_get_stopval(self)

    def set_stopval(self, stopval):
        return _nlopt.opt_set_stopval(self, stopval)

    def get_ftol_rel(self):
        return _nlopt.opt_get_ftol_rel(self)

    def set_ftol_rel(self, ftol_rel):
        return _nlopt.opt_set_ftol_rel(self, ftol_rel)

    def get_ftol_abs(self):
        return _nlopt.opt_get_ftol_abs(self)

    def set_ftol_abs(self, ftol_abs):
        return _nlopt.opt_set_ftol_abs(self, ftol_abs)

    def get_xtol_rel(self):
        return _nlopt.opt_get_xtol_rel(self)

    def set_xtol_rel(self, xtol_rel):
        return _nlopt.opt_set_xtol_rel(self, xtol_rel)

    def get_xtol_abs(self, *args):
        return _nlopt.opt_get_xtol_abs(self, *args)

    def set_xtol_abs(self, *args):
        return _nlopt.opt_set_xtol_abs(self, *args)

    def get_maxeval(self):
        return _nlopt.opt_get_maxeval(self)

    def set_maxeval(self, maxeval):
        return _nlopt.opt_set_maxeval(self, maxeval)

    def get_numevals(self):
        return _nlopt.opt_get_numevals(self)

    def get_maxtime(self):
        return _nlopt.opt_get_maxtime(self)

    def set_maxtime(self, maxtime):
        return _nlopt.opt_set_maxtime(self, maxtime)

    def get_force_stop(self):
        return _nlopt.opt_get_force_stop(self)

    def set_force_stop(self, force_stop):
        return _nlopt.opt_set_force_stop(self, force_stop)

    def force_stop(self):
        return _nlopt.opt_force_stop(self)

    def get_errmsg(self):
        return _nlopt.opt_get_errmsg(self)

    def set_local_optimizer(self, lo):
        return _nlopt.opt_set_local_optimizer(self, lo)

    def get_population(self):
        return _nlopt.opt_get_population(self)

    def set_population(self, population):
        return _nlopt.opt_set_population(self, population)

    def get_vector_storage(self):
        return _nlopt.opt_get_vector_storage(self)

    def set_vector_storage(self, vector_storage):
        return _nlopt.opt_set_vector_storage(self, vector_storage)

    def set_initial_step(self, *args):
        return _nlopt.opt_set_initial_step(self, *args)

    def set_default_initial_step(self, x):
        return _nlopt.opt_set_default_initial_step(self, x)

    def get_initial_step(self, *args):
        return _nlopt.opt_get_initial_step(self, *args)

    def get_initial_step_(self, x):
        return _nlopt.opt_get_initial_step_(self, x)
opt_swigregister = _nlopt.opt_swigregister
opt_swigregister(opt)


def srand(seed):
    return _nlopt.srand(seed)
srand = _nlopt.srand

def srand_time():
    return _nlopt.srand_time()
srand_time = _nlopt.srand_time

def version(major, minor, bugfix):
    return _nlopt.version(major, minor, bugfix)
version = _nlopt.version

def version_major():
    return _nlopt.version_major()
version_major = _nlopt.version_major

def version_minor():
    return _nlopt.version_minor()
version_minor = _nlopt.version_minor

def version_bugfix():
    return _nlopt.version_bugfix()
version_bugfix = _nlopt.version_bugfix

def algorithm_name(a):
    return _nlopt.algorithm_name(a)
algorithm_name = _nlopt.algorithm_name
# This file is compatible with both classic and new-style classes.

_algorithm = [GN_DIRECT_L, GN_DIRECT_L_RAND, GN_MLSL, GN_CRS2_LM, GN_ISRES, GN_ESCH, LN_BOBYQA, LN_COBYLA, LN_NELDERMEAD, LN_SBPLX,  LN_PRAXIS, LN_AUGLAG_EQ]

class NLopt(Algorithm):
    """ NLopt algorithms """

    def __init__(self, problem: Problem, name="NLopt"):
        super().__init__(problem, name)

        self.problem.options['save_level'] = 'individual'
        self.job = JobSimple(self.problem)
        self.options.declare(name='algorithm', default=LN_BOBYQA, values=_algorithm,
                             desc='Algorithm')
        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum evaluations')
        self.options.declare(name='xtol_rel', default=1e-6, lower=0.0,
                             desc='xtol_rel')
        self.options.declare(name='xtol_abs', default=1e-12, lower=0.0,
                             desc='xtol_abs')
        self.options.declare(name='ftol_rel', default=1e-6, lower=0.0,
                             desc='ftol_rel')
        self.options.declare(name='ftol_abs', default=1e-12, lower=0.0,
                             desc='ftol_abs')

    def _function(self, x, grad):
        return self.job.evaluate_scalar(x)

    def _constraint(x, grad, a, b):
        # if grad.size > 0:
        #     grad[0] = 3 * a * (a * x[0] + b) ** 2
        #     grad[1] = -1.0
        # return (a * x[0] + b) ** 3 - x[1]
        return 0

    def run(self):
        # Figure out bounds vectors.
        lb = []
        ub = []
        for parameter in self.problem.parameters:
            bounds = parameter['bounds']

            lb.append(bounds[0])
            ub.append(bounds[1])

        op = opt(self.options['algorithm'], len(self.problem.parameters))
        op.set_lower_bounds(lb)
        op.set_upper_bounds(ub)
        op.set_min_objective(self._function)
        op.set_xtol_rel(self.options['xtol_rel'])
        op.set_xtol_abs(self.options['xtol_abs'])
        op.set_ftol_rel(self.options['ftol_rel'])
        op.set_ftol_abs(self.options['ftol_abs'])
        op.set_maxeval(self.options['n_iterations'])

        # constraint
        # op.add_inequality_constraint(lambda x, grad: myconstraint(x, grad, 2, 0), 1e-8)
        # op.add_inequality_constraint(lambda x, grad: myconstraint(x, grad, -1, 1), 1e-8)

        try:
            t_s = time.time()
            self.problem.logger.info("NLopt: {}".format(op.get_algorithm_name()))
            x = op.optimize(self.problem.get_initial_values())
            # print('initial values:',x)
            t = time.time() - t_s
            self.problem.logger.info("NLopt: elapsed time: {} s".format(t))

            """
            if self.options['verbose_level'] >= 1:
                print('method: ', op.get_algorithm_name())
                print('optimum at ', x)
                print('minimum value = ', op.last_optimum_value())
                print('nevals = ', op.get_numevals())
            """
        except RuntimeError:
            print('Optimization FAILED.')
            print(op.get_errmsg())
        except ValueError:
            print('Optimization FAILED.')
            print(op.get_errmsg())

        msg_nlopt = {-1: 'failure - generic failure code',
                     -2: 'failure - invalid arguments',
                     -3: 'failure - out of memory',
                     -4: 'failure - round off limited',
                     -5: 'failure - forced stop',
                      1: 'success - generic success code',
                      2: 'success - stop value reached',
                      3: 'success - ftol reached',
                      4: 'success - xtol reached',
                      5: 'success - maxeval reached',
                      6: 'success - maxtime reached'
                     }

        if self.options['verbose_level'] >= 1:
            print('optimum = ', op.last_optimum_value())
            print('result code and meaning = ', op.last_optimize_result(), msg_nlopt[op.last_optimize_result()])


# import os
# import unittest
#
# from artap.executor import RemoteSSHExecutor
# from artap.problem import Problem
#
#
# class RemoteProblem(Problem):
#     """ Describe simple one objective optimization problem. """
#     def __init__(self, name):
#         parameters = {'x_1': {'initial_value': 10},
#                       'x_2': {'initial_value': 10}}
#         costs = ['F1']
#
#         super().__init__(name, parameters, costs,
#                          working_dir="." + os.sep + "workspace" + os.sep + "remote")
#
#     def evaluate(self, x):
#         return self.executor.eval(x)
#
#     def parse_results(self, content):
#         return [float(content)]
#
#
# class TestRemoteSSHExecutor(unittest.TestCase):
#     """ Tests simple optimization problem where calculation of
#         goal function is performed on remote machine.
#     """
#     def test_remote_python_exec(self):
#         problem = RemoteProblem("TestPythonProblem")
#         problem.executor = RemoteSSHExecutor(problem,
#                                              command="python3",
#                                              model_file="run_exec.py",
#                                              output_file="output.txt")
#
#         result = problem.evaluate([1, 2])
#
#         self.assertAlmostEqual(result[0], 5.0)
#
#     def test_remote_python_input(self):
#         problem = RemoteProblem("TestPythonProblem")
#         problem.executor = RemoteSSHExecutor(problem,
#                                              command="python3",
#                                              model_file="run_input.py",
#                                              input_file="input.txt",  # file is created in eval with specific parameters
#                                              output_file="output.txt")
#
#         result = problem.evaluate([1, 2])
#
#         self.assertAlmostEqual(result[0], 5.0)
#
#     def xtest_remote_octave_exec(self):
#         problem = RemoteProblem("TestOctaveProblem")
#         problem.executor = RemoteSSHExecutor(problem,
#                                              command="octave --no-gui",
#                                              model_file="run_input.m",
#                                              input_file="input.txt", # file is created in eval with specific parameters
#                                              output_file="output.txt")
#
#         result = problem.evaluate([1, 2])
#
#         self.assertAlmostEqual(result[0], 5.0)
#
#     def xtest_remote_matlab_run(self):
#         problem = RemoteProblem("TestMatlabProblem")
#
#         # TODO: for matlab it must be without .m extension
#         problem.executor = RemoteSSHExecutor(problem,
#                                              command="/opt/matlab-R2018b/bin/matlab -nodisplay -nosplash -nodesktop -r",
#                                              model_file="run_input.m",
#                                              input_file="input.txt",  # file is created in eval with specific parameters
#                                              output_file="output.txt")
#
#         result = problem.evaluate([])
#
#         self.assertAlmostEqual(result[0], 5.0)
#
#
# if __name__ == '__main__':
#     unittest.main()



import unittest

from artap.problem import Problem
from artap.benchmark_functions import BinhAndKorn
from artap.algorithm_genetic import NSGAII
from artap.results import Results

import optproblems as optp
import optproblems.cec2005 as cec2005


class BinhAndKornProblem(Problem):
    """
    Describe simple one objective optimization problem.
    Note: This test is not removed, because its not part of the cec2005 library.
    """

    def set(self):
        self.name = "TestNSGA2Optimization"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [0, 3]}]
        self.costs = [{'name': 'F_1', 'criteria': 'minimize'}, {'name': 'F_2', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        function = BinhAndKorn()
        return function.eval(individual.vector)

    def evaluate_constraints(self, individual):
        return BinhAndKorn.constraints(individual.vector)


class TestNSGA2Optimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem_nsga2(self):

        problem = BinhAndKornProblem()

        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 50
        algorithm.options['max_population_size'] = 200
        algorithm.options['calculate_gradients'] = True
        algorithm.options['verbose_level'] = 1

        algorithm.run()
        b = Results(problem)
        solution = b.pareto_values()
        wrong = 0

        for sol in solution:
            if abs(BinhAndKorn.approx(sol[0]) - sol[1]) > 0.1 * BinhAndKorn.approx(sol[0]) \
                    and 20 < sol[0] < 70:
                wrong += 1

        self.assertLessEqual(wrong, 5)

    def test_local_problem_nsga2(self, population_number=5):
        try:
            problem = BinhAndKornProblem()
            algorithm = NSGAII(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 50
            algorithm.options['calculate_gradients'] = True
            algorithm.options['verbose_level'] = 1

            algorithm.run()
            b = Results(problem)
            solution = b.pareto_values()
            wrong = 0

            # c = GraphicalResults(problem)
            # c.plot_populations()

            for sol in solution:
                if abs(BinhAndKorn.approx(sol[0]) - sol[1]) > 0.1 * BinhAndKorn.approx(sol[0]) \
                        and 20 < sol[0] < 70:
                    wrong += 1
            print(wrong)
            self.assertLessEqual(wrong, 5)
        except AssertionError:
            # stochastic
            print("TestNSGA2Optimization::test_local_problem_nsga2", population_number)
            self.test_local_problem_nsga2(int(1.5 * population_number))


class CEC2005_TEST_Problems(Problem):
    """
    Template class to initialize the different test problems these problems working for 2, 10, 30 variables.

    There are 25 test problems, which were compiled for the Special Session on Real-Parameter Optimization at the
    Congress on Evolutionary Computation (CEC), Edinburgh, UK, 2-5 Sept. 2005. The mathematical definitions are given in
    [Suganthan2005]. Additionally, implementations in C, Java, and Matlab were provided for the participants.
    Artap uses the optproblems test library to run these test. This template class is made to invoke them.
    """

    def set(self):
        # two variables
        self.parameters = [{'name': 'x_1', 'bounds': [func.min_bounds[0], func.max_bounds[0]]},
                           {'name': 'x_2', 'bounds': [func.min_bounds[1], func.max_bounds[1]]}]

        self.costs = [{'name':'F_1'}]


    def evaluate(self, x):
        problem = optp.Problem(func, num_objectives=1)
        solutions = [optp.Individual(x.vector)]

        problem.batch_evaluate(solutions)

        return [solutions[0].objective_values]


class TestCEC2005(unittest.TestCase):
    """Tests from CEC2005 library"""

    def run_test_problem(self, nr_gen, function):
        global func
        func = function
        problem = CEC2005_TEST_Problems()
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = nr_gen
        algorithm.options['max_population_size'] = 100
        algorithm.run()

        b = Results(problem)
        optimum = b.find_minimum('F_1')  # Takes last cost function
        self.assertAlmostEqual(optimum.costs[0], func.bias, 0)
        del func

    def test_shifted_sphere(self):
        self.run_test_problem(50, optp.cec2005.F1(2))

    def test_shifted_double_sum(self):
        self.run_test_problem(50, optp.cec2005.F2(2))

    def test_schwefel(self):  # !
        self.run_test_problem(100, optp.cec2005.F5(2))

    def test_shifted_rosenbrock(self):
        self.run_test_problem(100, optp.cec2005.F6(2))

    def test_shifted_rastrigin(self):
        self.run_test_problem(100, optp.cec2005.F9(2))

    def test_shifted_rot_weierstrass(self):
        self.run_test_problem(100, optp.cec2005.F11(2))

    def test_fletcherpowell(self):  # schwefel
        self.run_test_problem(100, optp.cec2005.F12(2))

    def test_f8f2(self):
        self.run_test_problem(100, optp.cec2005.F13(2))

    def test_expanded_f6(self):
        self.run_test_problem(100, optp.cec2005.F14(2))

    def test_hybrid_composition(self):
        self.run_test_problem(100, optp.cec2005.F15(2))

    def test_rotated_f15(self):
        self.run_test_problem(100, optp.cec2005.F16(2))

    def test_f17_with_noise(self):
        self.run_test_problem(250, optp.cec2005.F17(2))


if __name__ == '__main__':
    unittest.main()



import unittest
from artap.dynamical_systems import DiscreteLTISystem
from artap.dynamical_systems import Model
from artap.dynamical_systems import Gain
from artap.dynamical_systems import ContinuousLTISystem
import pylab as pl


class TestDynamicalSystems(unittest.TestCase):

    def test_discrete_LTI(self):

        matrix_a = [[-1, 0],
                    [0, -1]]

        matrix_c = [[1, 0], [0, 1]]

        matrix_b = [[1], [2]]
        matrix_d = [0]

        plant = Model(dt=0.1)
        x_0 = [10, 5]
        system = DiscreteLTISystem(plant, matrix_a, matrix_b, matrix_c, matrix_d, x_0, dt=1)
        plant.run(10)

        self.assertEqual(len(system.archive_t), len(system.archive_u))
        self.assertEqual(len(system.archive_t), len(system.archive_x))
        self.assertEqual(len(system.archive_t), len(system.archive_y))
        self.assertEqual(system.archive_x[0][0], x_0[0])
        self.assertEqual(system.archive_x[0][1], x_0[1])
        self.assertAlmostEqual(system.archive_x[-2][0], -10)
        self.assertAlmostEqual(system.archive_x[-2][1], -5)

    def test_continuous_LTI(self):
        plant = Model(dt=0.1)
        A = [[-1,  -1],
             [1,  0]]

        C = [0,  1]

        B = [[1], [1]]
        D = [0]

        plant = Model(dt=0.01)
        x0 = [10, 5]
        system = ContinuousLTISystem(plant, A, B, C, D, x0)
        plant.run(10)
        system.plot('output')
        pl.show()


# A = [[-1, 0],
#      [0, -1]]
#
# C = [[1, 0], [0, 1]]
#
# B = [[1], [2]]
# D = [0]
#
# plant = Model(dt=0.1)
# x0 = [10, 5]
# system = DiscreteLTISystem(plant, A, B, C, D, x0, dt=1)
# system.connect(unit_step)
# plant.run(10)
# system.plot('output')
# pl.show()


# plant = Model(dt=0.1)
# x0 = [4000, 280]
# A = [[1,  1],
#      [0,  1]]
#
# C = [[1,  0], [0,  1]]
#
# B = [[1], [2]]
# D = [0]
#
#
# def train_sequence(time):
#
#     y = np.array([4000, 280])
#     if time == 1:
#         y = np.array([4260, 282])
#     if time == 2:
#         y = np.array([4550, 285])
#     if time == 3:
#         y = np.array([4860, 286])
#     if time == 4:
#         y = np.array([5110, 290])
#     return y
#
#
# error_obs = [25, 6]
# error_proc = [20, 5]
#
# system = DiscreteLTISystem(plant, A, B, C, D, x_0=x0, dt=1)
# kalman_filter = KalmanFilter(plant, system, dt=1, error_obs=error_obs, error_proc=error_proc)
#
# system.connect(unit_step)
# kalman_filter.connect(unit_step)
# kalman_filter.connect(train_sequence, 1, 1)
#
# plant.run(3)
# kalman_filter.plot('output', index=1)
# pl.show()


# plant = Model(0.1)
# pi = PI(plant, 1, 0.015)
# pi.connect(unit_step)
# sumator = Sum(plant, [1, 1])
# sumator.connect(pi.output, output_index=0, input_index=0)
# sumator.connect(unit_step, output_index=0, input_index=1)
# plant.run(10)
# sumator.plot('output')
# pl.show()


# A = [[-1,  -1],
#      [0,  -1]]
#
# C = [[1,  0], [0,  1]]
#
# B = [[1, 1], [2, 2]]
# D = [0, 0]
#
# plant = Model(dt=0.01)
# x0 = [0, 1]
# system = ContinuousLTISystem(plant, A, B, C, D, x0)
# plant.run(100)
# system.plot('state')
# pl.show()

# x0 = [0, 0, 0]
# A = [[-1,  -1,  0],
#      [1,  0, -1],
#      [0,  1, 0]]
#
# C = [[1,  0,  0]]
#
# B = [[1], [-1], [2]]
# D = [0]
#
# plant = Model(dt=0.01)
# system = ContinuousLTISystem(plant, A, B, C, D, x0)
# system.connect(unit_step, output_index=0, input_index=0)
# plant.run(20)
# system.plot('output')
# pl.show()
# T = system.observability_matrix()
# T_1 = np.linalg.inv(T)
# A_ = T @ A @ T_1
# B_ = T @ B
# C_ = C @ T
# x0_ = T @ x0
# print(A_)
# print(B_)
# print(C_)
#
# plant = Model(dt=0.01)
# system_ = ContinuousLTISystem(plant, A_, B_, C_, D, x0_)
# system_.connect(unit_step, output_index=0, input_index=0)
# plant.run(20)
# system_.plot('output')
# pl.show()
# print((A_[-1]))
# nom = list(np.array(B_)[:, 0])
# nom.reverse()
# print(nom)
# print(np.roots(nom))


# A = [[0.596,   -0.214,   -0.051,    0.002],
#      [-0.435,    0.642,  -0.093,    0.004],
#      [-0.152,   -0.166,   0.933,    0.004],
#      [0.278,    0.350,    0.109,    0.994]]
#
# B = [[-60.846],
#      [-107.441],
#      [-83.854],
#      [217.776]]
#
# C = [[.0287,    0.0081,    0.0181,    0.0208]]
#
# D = [0]
#
# x0 = [0, 0, 0, 0]
#
# error_obs = [1]
# error_proc = [100, 100, 100, 100]
#
# plant = Model(dt=0.01)
# system = DiscreteLTISystem(plant, A, B, C, D, x0, dt=0.1)
# system.connect(programme_control)
# suma = Sum(plant, signs=[1, 1])
# suma.connect(system.output, input_index=1, output_index=0)
# suma.connect(white_noise, input_index=0, output_index=0)
# kalman_filter = KalmanFilter(plant, system, dt=0.1, error_obs=error_obs, error_proc=error_proc)
# kalman_filter.connect(programme_control)
# kalman_filter.connect(suma.output, input_index=1, output_index=0)
# plant.run(400)
# suma.plot('output')
# pl.figure()
# kalman_filter.plot('output')
# system.plot('output')
# pl.show()

# dt = 0.2
#
# A = [[1,   0,   0,    0],
#      [dt,  1,   0,    0],
#      [0,   0,   1,    0],
#      [0,   0,   dt,   1]]
#
# B = [[0],
#      [0],
#      [-9.8],
#      [0]]
#
# C = [[1,    0,    0,    0],
#      [0,    1,    0,    0],
#      [0,    0,    1,    0],
#      [0,    0,    0,    1]]
#
# D = [[0], [0], [0], [0]]
#
# x0 = [-600, 3e3, 100, 600]
#
# error_obs = [10, 10, 10, 10]
# error_proc = [200, 20, 20, 20]
#
# plant = Model(dt=0.01)
# system = DiscreteLTISystem(plant, A, B, C, D, x0, dt=0.2)
# system.connect(unit_step)
#
# suma = Sum(plant, signs=[1.0, 1.0])
# # suma.connect(system.output, input_index=0, output_index=0)
# suma.connect(white_noise, input_index=0, output_index=0)
# kalman_filter = KalmanFilter(plant, system, dt=0.2, error_obs=error_obs, error_proc=error_proc)
#
# kalman_filter.connect(unit_step)
# kalman_filter.connect(system.output, input_index=1, output_index=0)
# plant.run(6.5)
# kalman_filter.plot('input')
#
# x = system.get_state()
# pl.plot(x[3])
# x_hat = kalman_filter.get_state()
# pl.plot(x_hat[3], 'x')
# pl.show()
# pl.figure(1)
# system.plot('output',index=0)
# pl.show()


if __name__ == '__main__':
    unittest.main()



from .algorithm import Algorithm
from .job import JobSimple
from .operators import RandomGeneration
from .operators import GradientGeneration
from .population import Population
import time


class SweepAlgorithm(Algorithm):
    """
    Sweep Analysis
    """

    def __init__(self, problem, generator=None, name='SweepAlgorithm'):
        super().__init__(problem, name)

        if generator is None:
            # set random generator
            self.generator = RandomGeneration(self.problem.parameters)
            self.generator.init(10)
        else:
            self.generator = generator

        self.job = JobSimple(self.problem)

    def run(self):
        t_s = time.time()
        self.problem.logger.info(
            "Sweep: {}: {} individuals".format(self.generator.__class__.__name__, self.population_size))

        # create initial population and evaluate individuals
        population = self.gen_initial_population()

        if self.options['calculate_gradients'] is True:
            self.problem.options['save_level'] = 'population'
            # TODO: Move to Algorithm class here only call calculate_gradients(individuals)
            gradient_generator = GradientGeneration(self.problem.parameters)
            gradient_generator.init(population.individuals)
            gradient_population = Population(gradient_generator.generate())
            gradient_population.individuals = self.evaluate(gradient_population.individuals)
            population.individuals = self.evaluate_gradient(population.individuals, gradient_population.individuals)

        if self.problem.options['save_level'] == 'population':
            self.problem.data_store.write_population(population)

        t = time.time() - t_s
        self.problem.logger.info("Sweep: elapsed time: {} s".format(t))



from .algorithm import *
from .executor import *
from .algorithm_scipy import *
from .algorithm_genetic import *

import platform

if platform.system() == 'Linux':
    from .algorithm_bayesopt import *

    __all__ = ['Problem', 'NSGA_II', 'BayesOpt']

else:
    __all__ = ['Problem', 'NSGA_II']
__all__ = ['Problem', 'NSGA_II']


from abc import ABCMeta, abstractmethod
from .utils import VectorAndNumbers
from .individual import Individual
from multiprocessing import Queue
import os

from copy import deepcopy


class Job(metaclass=ABCMeta):
    def __init__(self, problem, population):
        self.problem = problem
        self.population = population

    @abstractmethod
    def evaluate(self, x):
        pass

    def evaluate_scalar(self, x):
        # simple individual
        individual = Individual(x)
        self.evaluate(individual)

        return individual.costs[0]


class JobSimple(Job):
    def __init__(self, problem, population=None):
        super().__init__(problem, population)

    def evaluate(self, individual):
        # check the constraints
        constraints = self.problem.evaluate_constraints(individual)

        if constraints:
            individual.feasible = sum(map(abs, constraints))

        # problem cost function evaluate only in that case when the problem is fits the constraints
        try:
            individual.costs = self.problem.surrogate.evaluate(individual)
            # self.problem.surrogate.evaluate(individual)

            individual.state = individual.State.EVALUATED

        except (TimeoutError, RuntimeError) as e:
            # individual.vector = VectorAndNumbers.gen_vector(self.problem.parameters)
            individual.state = Individual.State.FAILED

        # add to population
        if self.population is not None:
            self.population.individuals.append(individual)

        # write to datastore
        if self.problem.options["save_level"] == "individual":
            self.problem.data_store.write_individual(individual)


class JobQueue(Job):
    def __init__(self, problem, shared_list, queue: Queue, population=None):
        super().__init__(problem, population)

        self.shared_list = shared_list
        self.queue = queue

    def evaluate(self, individual):
        # global individual
        for item in self.shared_list:
            if item[0] == os.getpid():
                individual_local = deepcopy(item[1])
        """
        if self.shared_list is not None:
            for item in self.shared_list:
                if item[0] == os.getpid():
                    individual = deepcopy(item[1])
        else:
            individual = Individual(x)
        """
        # check the constraints
        constraints = self.problem.evaluate_constraints(individual)

        if constraints:
            individual_local.feasible = sum(map(abs, constraints))

        # problem cost function evaluate only in that case when the problem is fits the constraints
        try:
            individual_local.costs = self.problem.surrogate.evaluate(individual)
            individual_local.state = Individual.State.EVALUATED
        except (TimeoutError, RuntimeError) as e:
            # individual_local.vector = VectorAndNumbers.gen_vector(self.problem.parameters)
            individual_local.state = Individual.State.FAILED

        if self.queue is not None:
            self.queue.put(individual_local)

        # add to population
        if self.population is not None:
            self.population.individuals.append(individual)



from .problem import Problem
from .algorithm import Algorithm
from .job import JobSimple


from scipy.optimize import minimize
import numpy as np
import time


_algorithm = ['Nelder-Mead', 'Powell', 'CG', 'BFGS', 'Newton-CG', 'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP', 'dogleg', 'trust-ncg', 'trust-exact', 'trust-krylov']


class ScipyOpt(Algorithm):
    """ Class is prepared for usage of Nelder-Mead method from package SciPy."""

    def __init__(self, problem: Problem, name="ScipyOpt"):
        super().__init__(problem, name)

        self.problem.options['save_level'] = 'individual'
        self.options.declare(name='algorithm', default='Nelder-Mead', values=_algorithm,
                             desc='Algorithm')
        self.options.declare(name='tol', default=1e-6, lower=0.0,
                             desc='tol')
        self.options.declare(name='bounds', default=None, desc='bounds')

        self.save_all = True

    def run(self):
        job = JobSimple(self.problem)

        # initial vector
        x0 = np.array(self.problem.get_initial_values())

        # optimization
        t_s = time.time()
        self.problem.logger.info("ScipyOpt: {}".format(self.options['algorithm']))
        minimize(job.evaluate_scalar, x0, method=self.options['algorithm'], tol=self.options['tol'], bounds=self.options['bounds'])
        t = time.time() - t_s
        self.problem.logger.info("ScipyOpt: elapsed time: {} s".format(t))



from abc import abstractmethod, ABC
import sys
import random
import math
import itertools

from .individual import Individual
from .utils import VectorAndNumbers
from .doe import build_box_behnken, build_lhs, build_full_fact, build_plackett_burman

EPSILON = sys.float_info.epsilon


class Operation(ABC):

    def __init__(self):
        pass

    @staticmethod
    def clip(value, min_value, max_value):
        return max(min_value, min(value, max_value))


class Generation(Operation):

    def __init__(self, problem=None, individual_class=Individual, parameters=None):
        super().__init__()
        if problem is not None:
            self.parameters = problem.parameters
        else:
            self.parameters = parameters
        self.individual_class = individual_class

    def create_individual(self, vector: list=[]):
        return self.individual_class(vector)

    @abstractmethod
    def generate(self):
        pass


class CustomGeneration(Generation):

    def __init__(self, problem=None):
        super().__init__(problem)
        self.vectors = []

    def init(self, vectors):
        self.vectors = vectors

    def generate(self):
        individuals = []
        for vector in self.vectors:
            individuals.append(self.create_individual(vector))
        return individuals


class RandomGeneration(Generation):

    def __init__(self, problem=None, individual_class=Individual, parameters=None):
        super().__init__(problem, individual_class, parameters=parameters)
        self.number = 0

    def init(self, number):
        self.number = number

    def generate(self):
        individuals = []
        for i in range(self.number):
            vector = VectorAndNumbers.gen_vector(self.parameters)
            individuals.append(self.create_individual(vector))
        return individuals


class FullFactorGeneration(Generation):
    """
    Create a general full-factorial design
    Number of experiments (2 ** len(parameters) - without center, 3 ** len(parameters - with center)
    """

    def __init__(self, problem=None, parameters=None):
        super().__init__(problem, parameters=parameters)
        self.center = False

    def init(self, center):
        self.center = center

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter['name']
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]

            if self.center:
                dict_vars[name] = [l_b, (l_b + u_b) / 2.0, u_b]
            else:
                dict_vars[name] = [l_b, u_b]

        df = build_full_fact(dict_vars)

        individuals = []
        for vector in df:
            individuals.append(self.create_individual(vector))
        return individuals


class PlackettBurmanGeneration(Generation):
    """
    Create a general full-factorial design
    Number of experiments (2 ** len(parameters) - without center, 3 ** len(parameters - with center)
    """

    def __init__(self, problem=None, parameters=None):
        super().__init__(problem, parameters=parameters)

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter['name']
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]
            dict_vars[name] = [l_b, u_b]

        df = build_plackett_burman(dict_vars)
        # print(df)

        individuals = []
        for vector in df:
            individuals.append(self.create_individual(vector))
        return individuals


class BoxBehnkenGeneration(Generation):
    """
    Create a general full-factorial design
    # 3 params = 13 experiments
    # 4 params = 25 experiments
    # 5 params = 41 experiments
    # 6 params = 49 experiments
    # 7 params = 57 experiments
    # 8 params = 113 experiments
    https://en.wikipedia.org/wiki/Box%E2%80%93Behnken_design
    """

    def __init__(self, problem=None, parameters=None):
        super().__init__(problem, parameters=parameters)

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter['name']
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]
            dict_vars[name] = [l_b, u_b]

        df = build_box_behnken(dict_vars)
        # print(df)

        individuals = []
        for vector in df:
            individuals.append(self.create_individual(vector))
        return individuals


class LHSGeneration(Generation):
    """
    Builds a Latin Hypercube design dataframe from a dictionary of factor/level ranges.
    """

    def __init__(self, problem=None, parameters=None):
        super().__init__(problem, parameters=parameters)
        self.number = 0

    def init(self, number):
        self.number = number

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter["name"]
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]
            dict_vars[name] = [l_b, u_b]

        df = build_lhs(dict_vars, num_samples=self.number)
        # print(df)

        individuals = []
        for vector in df:
            individuals.append(self.create_individual(vector))
        return individuals


class GradientGeneration(Generation):

    def __init__(self, problem=None, parameters=None):
        super().__init__(parameters=parameters)
        self.delta = 1e-6
        self.individuals = None

    def init(self, individuals):
        self.individuals = individuals

    def generate(self):
        new_individuals = []
        k = 0
        for individual in self.individuals:
            for i in range(len(individual.vector)):
                vector = individual.vector.copy()
                vector[i] -= self.delta
                new_individuals.append(self.create_individual(vector))
                new_individuals[-1].depends_on = k
                new_individuals[-1].modified_param = i
            k += 1
        return new_individuals


class Mutation(Operation):

    def __init__(self, parameters, probability):
        super().__init__()
        self.parameters = parameters
        self.probability = probability

    @abstractmethod
    def mutate(self, p):
        pass


class SimpleMutation(Mutation):
    def __init__(self, parameters, probability):
        super().__init__(parameters, probability)

    def mutate(self, p):
        """ uniform random mutation """
        mutation_space = 0.1
        vector = []

        for i, parameter in enumerate(self.parameters):
            if random.uniform(0, 1) < self.probability:
                l_b = parameter['bounds'][0]
                u_b = parameter['bounds'][1]

                para_range = mutation_space * (u_b - l_b)
                mutation = random.uniform(-para_range, para_range)
                # vector.append(self.clip(p.vector[i] + mutation, l_b, u_b))
                vector.append(p.vector[i] + mutation)
            else:
                vector.append(p.vector[i])

        p_new = p.__class__(vector)
        return p_new


class PmMutation(Mutation):
    """
    PmMutation -- for nsga2 and epsMoEA

    This operator can handle real, integer and boolean optimization parameters.
    The class contains two-kind of operators as the original (Deb's [...]) implementation .

    This is a difference between Artap and Platypus, where the Integer numbers are encoded (Gray-encoding) and handled
    as binary numbers with the Bitflip operator.
    """

    def __init__(self, parameters, probability, distribution_index=20):
        super().__init__(parameters, probability)
        self.distribution_index = distribution_index

    def mutate(self, parent):
        vector = []

        for i, parameter in enumerate(self.parameters):
            if isinstance(parent.vector[i], list):
                vector.append(self.bitflip(parent.vector[i]))

            else:
                if random.uniform(0, 1) < self.probability:
                    l_b = parameter['bounds'][0]
                    u_b = parameter['bounds'][1]

                    if isinstance(parent.vector[i], float):
                        vector.append(self.pm_mutation(parent.vector[i], l_b, u_b))

                    if isinstance(parent.vector[i], int):
                        vector.append(int(self.pm_mutation(parent.vector[i], l_b, u_b)))
                else:
                    vector.append(parent.vector[i])

        p_new = parent.__class__(vector)
        return p_new

    def pm_mutation(self, x, lb, ub):
        """
        Polynomial mutation for float and integer parameters.

        :param x: represents one parameter of the problem
        :param lb: lower bound
        :param ub: upper bound
        :return:
        """
        u = random.uniform(0, 1)
        dx = ub - lb

        if u < 0.5:
            bl = (x - lb) / dx
            b = 2.0 * u + (1.0 - 2.0 * u) * pow(1.0 - bl, self.distribution_index + 1.0)
            delta = pow(b, 1.0 / (self.distribution_index + 1.0)) - 1.0
        else:
            bu = (ub - x) / dx
            b = 2.0 * (1.0 - u) + 2.0 * (u - 0.5) * pow(1.0 - bu, self.distribution_index + 1.0)
            delta = 1.0 - pow(b, 1.0 / (self.distribution_index + 1.0))

        x = x + delta * dx
        x = self.clip(x, lb, ub)

        return x

    def bitflip(self, x:list):

        for j in range(1,len(x)):
            if random.uniform(0.0, 1.0) <= self.probability:
                x[j] = not x[j]

        return x


class SwarmMutation(Mutation):

    def __init__(self, parameters, probability=1):
        super().__init__(parameters, probability)
        self.w = 0.1  # constant inertia weight (how much to weigh the previous velocity)
        self.c1 = 2  # cognitive constant
        self.c2 = 1  # social constant
        self.best_individual = None

    # evaluate current fitness
    def evaluate_pso(self, individual):

        dominates = True

        for i in range(len(individual.best_costs)):
            if individual.costs[i] > individual.best_costs[i]:
                dominates = False

        # check to see if the current position is an individual best
        if dominates:
            individual.best_vector = individual.vector
            individual.best_costs = individual.costs

    # update new particle velocity
    def update_velocity(self, individual):

        for i in range(0, len(individual.vector)):

            r1 = 0.1 * random.random()
            r2 = 0.1 * random.random()

            vel_cognitive = self.c1 * r1 * (individual.best_vector[i] - individual.vector[i])
            vel_social = self.c2 * r2 * (self.best_individual.vector[i] - individual.vector[i])
            individual.velocity_i[i] = self.w * individual.velocity_i[i] + vel_cognitive + vel_social

    # update the particle position based off new velocity updates
    def update_position(self, individual):

        for parameter, i in zip(self.parameters, range(len(individual.vector))):
            individual.vector[i] = individual.vector[i] + individual.velocity_i[i]

            # adjust maximum position if necessary
            if individual.vector[i] > parameter['bounds'][1]:
                individual.vector[i] = parameter['bounds'][1]

            # adjust minimum position if necessary
            if individual.vector[i] < parameter['bounds'][0]:
                individual.vector[i] = parameter['bounds'][0]

    def update(self, best_individual):
        self.best_individual = best_individual

    def mutate(self, p):
        self.update_velocity(p)
        self.update_position(p)
        return p


class Selection(Operation):

    def __init__(self, parameters, signs, part_num=2):
        super().__init__()
        self.parameters = parameters
        self.signs = signs
        self.part_num = part_num
        self.comparator = ParetoDominance()

    @abstractmethod
    def select(self, population):
        pass

    def is_dominate(self, p, q):
        """
        :param p: current solution
        :param q: candidate
        :return: True if the candidate is better than the current solution
        """
        dominate = False

        # The cost function can be a float or a list of floats
        for i in range(0, len(p.costs)):
            if p.costs[i] > q.costs[i]:
                return False
            if p.costs[i] <= q.costs[i]:
                dominate = True
        return dominate

    def non_dominated_sort(self, population):
        pareto_front = []
        front_number = 1

        for p in population:
            p.domination_counter = 0
            p.front_number = None
            p.dominate = set()

            for q in population:
                if p is q:
                    continue
                if self.comparator.compare(p, q) == 1:          # TODO: simplify
                    p.dominate.add(q)
                elif self.comparator.compare(q, p) == 1:
                    p.domination_counter = p.domination_counter + 1

            if p.domination_counter == 0:
                p.front_number = front_number
                pareto_front.append(p)

        while not len(pareto_front) == 0:
            front_number += 1
            temp_set = []
            for p in pareto_front:
                for q in p.dominate:
                    q.domination_counter -= 1
                    if q.domination_counter == 0 and q.front_number is None:
                        q.front_number = front_number
                        temp_set.append(q)
            pareto_front = temp_set

    def sort_by_coordinate(self, population, dim):
        population.sort(key=lambda x: x.costs[dim])
        return population

    def crowding_distance(self, population):
        infinite = float("inf")
        n = len(population[0].costs)
        for dim in range(0, n):
            new_list = self.sort_by_coordinate(population, dim)

            new_list[0].crowding_distance += infinite
            new_list[-1].crowding_distance += infinite
            max_distance = new_list[0].vector[dim] - new_list[-1].vector[dim]
            for i in range(1, len(new_list) - 1):
                distance = new_list[i - 1].vector[dim] - new_list[i + 1].vector[dim]
                if max_distance == 0:
                    new_list[i].crowding_distance = 0
                else:
                    new_list[i].crowding_distance += distance / max_distance

        for p in population:
            p.crowding_distance = p.crowding_distance / n


class DummySelection(Selection):

    def __init__(self, parameters, part_num=2):
        super().__init__(parameters, part_num)

    def select(self, individuals):
        selection = []
        for individual in individuals:

            candidate = individual.__class__(individual.vector)
            candidate.costs = individual.costs
            candidate.front_number = individual.front_number
            candidate.best_vector = individual.best_vector
            candidate.best_costs = individual.best_costs

            selection.append(candidate)

        return selection


class Dominance(ABC):
    def __init__(self):
        pass

    def compare(self, p, q):
        raise NotImplementedError("method not implemented")


class EpsilonDominance(Dominance):
    """
    Epsilon dominance.

        Similar to Pareto dominance except if the two solutions are contained
        within the same epsilon-box, the solution closer to the optimal corner
        or the box is preferred.
    """

    def __init__(self, epsilons, signs=None):
        super(EpsilonDominance, self).__init__()

        if hasattr(epsilons, "__getitem__"):
            self.epsilons = epsilons
        else:
            self.epsilons = [epsilons]

    def same_box(self, p, q):

        # first check constraint violation
        if p.feasible != q.feasible:
            if p.feasible == 0:
                return False
            elif q.feasible == 0:
                return False
            elif p.feasible < q.feasible:
                return False
            elif q.feasible < p.feasible:
                return False

        # then use epsilon dominance on the objectives
        dominate1 = False
        dominate2 = False

        for i in range(len(p.costs)):
            o1 = p.costs[i]
            o2 = q.costs[i]

            # in artap we cannot change the direction of the optimization in this way
            # if problem.directions[i] == Problem.MAXIMIZE:
            #    o1 = -o1
            #    o2 = -o2

            epsilon = float(self.epsilons[i % len(self.epsilons)])
            i1 = math.floor(o1 / epsilon)
            i2 = math.floor(o2 / epsilon)

            if i1 < i2:
                dominate1 = True

                if dominate2:
                    return False
            elif i1 > i2:
                dominate2 = True

                if dominate1:
                    return False

        if not dominate1 and not dominate2:
            return True
        else:
            return False

    def compare(self, p, q):

        # first check constraint violation
        if p.feasible != q.feasible:
            if p.feasible == 0:
                return 2
            elif q.feasible == 0:
                return 1
            elif p.feasible < q.feasible:
                return 2
            elif q.feasible < p.feasible:
                return 1

        # then use epsilon dominance on the objectives
        dominate1 = False
        dominate2 = False

        for i in range(len(p.costs)):
            o1 = p.costs[i]
            o2 = q.costs[i]

            epsilon = float(self.epsilons[i % len(self.epsilons)])
            i1 = math.floor(o1 / epsilon)
            i2 = math.floor(o2 / epsilon)

            if i1 < i2:
                dominate1 = True

                if dominate2:
                    return 0
            elif i1 > i2:
                dominate2 = True

                if dominate1:
                    return 0

        if not dominate1 and not dominate2:
            dist1 = 0.0
            dist2 = 0.0

            for i in range(len(p.costs)):
                o1 = p.costs[i]
                o2 = q.costs[i]

                epsilon = float(self.epsilons[i % len(self.epsilons)])
                i1 = math.floor(o1 / epsilon)
                i2 = math.floor(o2 / epsilon)

                dist1 += math.pow(o1 - i1 * epsilon, 2.0)
                dist2 += math.pow(o2 - i2 * epsilon, 2.0)

            if dist1 < dist2:
                return -1
            else:
                return 1
        elif dominate1:
            return -1
        else:
            return 1


class ParetoDominance(Dominance):
    # from platypus core
    """Pareto dominance with constraints.

    If either solution violates constraints, then the solution with a smaller
    constraint violation is preferred.  If both solutions are feasible, then
    Pareto dominance is used to select the preferred solution.
    """

    def __init__(self, signs, epsilons=None):
        super(ParetoDominance, self).__init__()
        self.signs = signs

    def compare(self, p, q):
        if len(p.costs) != len(q.costs):
            print("Len {} - {}".format(len(p.costs), len(p.costs)))
        # first check constraint violation
        if p.feasible != q.feasible:
            if p.feasible == 0:
                return 1
            elif q.feasible == 0:
                return 2
            elif p.feasible < q.feasible:
                return 1
            elif q.feasible < p.feasible:
                return 2

        # The cost function can be a float or a list of floats
        # Check the pareto dominance on every value of the calculated vectors
        dominate_p = False
        dominate_q = False
        for i in range(0, len(p.costs)):
            sign = self.signs[i]
            if sign * p.costs[i] > sign * q.costs[i]:
                dominate_q = True

                if dominate_p:
                    return 0

            if sign * p.costs[i] < sign * q.costs[i]:
                dominate_p = True

                if dominate_q:
                    return 0

        if dominate_q == dominate_p:
            return 0
        elif dominate_p:
            return 1
        else:
            return 2


class Selection(Operation):

    def __init__(self, parameters, signs=None, part_num=2, dominance=ParetoDominance):
        super().__init__()
        self.parameters = parameters
        self.part_num = part_num
        self.comparator = dominance(signs=signs)  # ParetoDominance()
        self.signs = signs

    @abstractmethod
    def select(self, population):
        pass

    def is_dominate(self, p, q):
        """
        :param p: current solution
        :param q: candidate
        :return: True if the candidate is better than the current solution
        """
        dominate = False

        # The cost function can be a float or a list of floats
        for i in range(0, len(p.costs)):
            if p.costs[i] > q.costs[i]:
                return False
            if p.costs[i] <= q.costs[i]:
                dominate = True
        return dominate

    def sorting(self, generation):
        """
        This shoring can be dominated or non-dominated.
        :param generation: means the list of the population, it can be list of the individuals or the archive.
        :return:
        """
        pareto_front = []
        front_number = 1

        for p in generation:
            p.domination_counter = 0
            p.front_number = None
            p.dominate = set()

            for q in generation:
                if p is q:
                    continue
                if self.comparator.compare(p, q) == 1:          # TODO: simplify
                    p.dominate.add(q)
                    p.dominate.add(q)
                elif self.comparator.compare(q, p) == 1:
                    p.domination_counter += 1

            if p.domination_counter == 0:
                p.front_number = front_number
                pareto_front.append(p)

        while not len(pareto_front) == 0:
            front_number += 1
            temp_set = []
            for p in pareto_front:
                for q in p.dominate:
                    q.domination_counter -= 1
                    if q.domination_counter == 0 and q.front_number is None:
                        q.front_number = front_number
                        temp_set.append(q)
            pareto_front = temp_set
        return

    def sort_by_coordinate(self, population, dim):
        population.sort(key=lambda x: x.costs[dim])
        return population

    def crowding_distance(self, population):
        infinite = float("inf")
        n = len(population[0].costs)
        for dim in range(0, n):
            new_list = self.sort_by_coordinate(population, dim)

            new_list[0].crowding_distance += infinite
            new_list[-1].crowding_distance += infinite
            max_distance = new_list[0].vector[dim] - new_list[-1].vector[dim]
            for i in range(1, len(new_list) - 1):
                distance = new_list[i - 1].vector[dim] - new_list[i + 1].vector[dim]
                if max_distance == 0:
                    new_list[i].crowding_distance = 0
                else:
                    new_list[i].crowding_distance += distance / max_distance

        for p in population:
            p.crowding_distance = p.crowding_distance / n


class DummySelection(Selection):

    def __init__(self, parameters, signs, part_num=2):
        super().__init__(parameters, signs, part_num)

    def select(self, individuals):
        selection = []
        for individual in individuals:

            candidate = individual.__class__(individual.vector)
            candidate.costs = individual.costs
            candidate.front_number = individual.front_number
            candidate.best_vector = individual.best_vector
            candidate.best_costs = individual.best_costs

            selection.append(candidate)
        return selection


class TournamentSelection(Selection):

    def __init__(self, parameters, signs=None, dominance=ParetoDominance, epsilons=None):
        super().__init__(parameters, signs=signs)
        self.dominance = dominance(signs=signs, epsilons=epsilons)
        self.signs = signs
        # self.dominance = EpsilonDominance([0.01])

    def select(self, population):
        """
        Binary tournament selection:
        An individual is selected in the rank is lesser than the other or if crowding distance is greater than the other
        """

        # participants = random.sample(population, 2)
        # return min(participants, key=lambda x: (x.front_number, -x.crowding_distance))

        winner = random.choice(population)

        for _ in range(self.part_num - 1):
            candidate = random.choice(population)
            flag = self.dominance.compare(winner, candidate)

            if flag > 0:
                winner = candidate

        return winner

        # participants = random.sample(parents, part_num)
        # return min(participants, key=lambda x: (x.front_number, -x.crowding_distance))


class Archive(object):
    """An archive only containing non-dominated solutions."""

    def __init__(self, signs, dominance=ParetoDominance):
        super(Archive, self).__init__()
        self._dominance = dominance(signs=signs)
        self._contents = []

    def add(self, solution):
        flags = [self._dominance.compare(solution, s) for s in self._contents]
        dominates = [x > 0 for x in flags]
        nondominated = [x == 0 for x in flags]

        if any(dominates):
            return False
        else:
            self._contents = list(itertools.compress(self._contents, nondominated)) + [solution]
            return True

    def append(self, solution):
        self.add(solution)

    def extend(self, solutions):
        for solution in solutions:
            self.append(solution)

    def remove(self, solution):
        try:
            self._contents.remove(solution)
            return True
        except ValueError:
            return False

    def __len__(self):
        return len(self._contents)

    def __getitem__(self, key):
        return self._contents[key]

    def __iadd__(self, other):
        if hasattr(other, "__iter__"):
            for o in other:
                self.add(o)
        else:
            self.add(other)

        return self

    def __iter__(self):
        return iter(self._contents)


class EpsilonBoxArchive(Archive):

    def __init__(self, epsilons):
        super(EpsilonBoxArchive, self).__init__(EpsilonDominance(epsilons))
        self.improvements = 0

    def add(self, solution):
        flags = [self._dominance.compare(solution, s) for s in self._contents]
        dominates = [x > 0 for x in flags]
        nondominated = [x == 0 for x in flags]
        dominated = [x < 0 for x in flags]
        not_same_box = [not self._dominance.same_box(solution, s) for s in self._contents]

        if any(dominates):
            return False
        else:
            self._contents = list(itertools.compress(self._contents, nondominated)) + [solution]

            if dominated and not_same_box:
                self.improvements += 1


class Crossover(Operation):

    def __init__(self, parameters, probability):
        super().__init__()
        self.parameters = parameters
        self.probability = probability

    @abstractmethod
    def cross(self, p1, p2):
        pass


class SimpleCrossover(Crossover):

    def __init__(self, parameters, probability):
        super().__init__(parameters, probability)

    def cross(self, p1, p2):
        parent_a = p1.vector.copy()
        parent_b = p2.vector.copy()

        if random.uniform(0.0, 1.0) <= self.probability:
            parameter1 = []
            parameter2 = []
            linear_range = 2

            alpha = random.uniform(0, linear_range)

            for i, param in enumerate(self.parameters):
                l_b = param['bounds'][0]
                u_b = param['bounds'][1]

                parameter1.append(self.clip(alpha * p1.vector[i] + (1 - alpha) * p2.vector[i], l_b, u_b))
                parameter2.append(self.clip((1 - alpha) * p1.vector[i] + alpha * p2.vector[i], l_b, u_b))

            parent_a = parameter1
            parent_b = parameter2

        offspring_a = p1.__class__(parent_a)
        offspring_b = p2.__class__(parent_b)

        return offspring_a, offspring_b


class SimulatedBinaryCrossover(Crossover):

    def __init__(self, parameters, probability, distribution_index=15):
        super().__init__(parameters, probability)
        self.distribution_index = distribution_index

    def sbx(self, x1, x2, lb, ub, p_type="real"):

        dx = x2 - x1

        if dx > EPSILON:
            if x2 > x1:
                y2 = x2
                y1 = x1
            else:
                y2 = x1
                y1 = x2

            beta = 1.0 / (1.0 + (2.0 * (y1 - lb) / (y2 - y1)))
            alpha = 2.0 - pow(beta, self.distribution_index + 1.0)
            rand = random.uniform(0.0, 1.0)

            if rand <= 1.0 / alpha:
                alpha = alpha * rand
                betaq = pow(alpha, 1.0 / (self.distribution_index + 1.0))
            else:
                alpha = alpha * rand
                alpha = 1.0 / (2.0 - alpha)
                betaq = pow(alpha, 1.0 / (self.distribution_index + 1.0))

            x1 = 0.5 * ((y1 + y2) - betaq * (y2 - y1))
            beta = 1.0 / (1.0 + (2.0 * (ub - y2) / (y2 - y1)))
            alpha = 2.0 - pow(beta, self.distribution_index + 1.0)

            if rand <= 1.0 / alpha:
                alpha = alpha * rand
                betaq = pow(alpha, 1.0 / (self.distribution_index + 1.0))
            else:
                alpha = alpha * rand
                alpha = 1.0 / (2.0 - alpha)
                betaq = pow(alpha, 1.0 / (self.distribution_index + 1.0))

            x2 = 0.5 * ((y1 + y2) + betaq * (y2 - y1))

            # randomly swap the values
            if bool(random.getrandbits(1)):
                x1, x2 = x2, x1

            x1 = self.clip(x1, lb, ub)
            x2 = self.clip(x2, lb, ub)

            if p_type == "integer":
                x1 = int(x1)
                x2 = int(x2)

        return x1, x2

    def cross(self, p1, p2):
        """
        Create an offspring using simulated binary crossover.
        :return:  a list with 2 offsprings each with the genotype of an  offspring after recombination and mutation.
        """

        parent_a = p1.vector.copy()
        parent_b = p2.vector.copy()

        if random.uniform(0.0, 1.0) <= self.probability:
            for i, param in enumerate(self.parameters):
                if random.uniform(0.0, 1.0) <= 0.5:
                    x1 = parent_a[i]
                    x2 = parent_b[i]

                    lb = param['bounds'][0]
                    ub = param['bounds'][1]

                    if not 'parameter_type' in param:
                        continue
                    else:
                        p_type = param['parameter_type']

                    x1, x2 = self.sbx(x1, x2, lb, ub, p_type)

                    parent_a[i] = x1
                    parent_b[i] = x2

        offspring_a = p1.__class__(parent_a)
        offspring_b = p2.__class__(parent_b)

        return offspring_a, offspring_b



import unittest
from artap.problem import Problem
from artap.algorithm_swarm import PSO
from artap.results import Results

import optproblems as optp
import optproblems.cec2005 as cec2005


class CEC2005_TEST_Problems(Problem):
    """
    Template class to initialize the different test problems these problems working for 2, 10, 30 variables.

    There are 25 test problems, which were compiled for the Special Session on Real-Parameter Optimization at the
    Congress on Evolutionary Computation (CEC), Edinburgh, UK, 2-5 Sept. 2005. The mathematical definitions are given in
    [Suganthan2005]. Additionally, implementations in C, Java, and Matlab were provided for the participants.
    Artap uses the optproblems test library to run these test. This template class is made to invoke them.
    """

    def set(self):
        # two variables
        self.parameters = [{'name': 'x_1', 'bounds': [func.min_bounds[0], func.max_bounds[0]]},
                           {'name': 'x_2', 'bounds': [func.min_bounds[1], func.max_bounds[1]]}]

        self.costs = [{'name':'F_1'}]


    def evaluate(self, x):
        problem = optp.Problem(func, num_objectives=1)
        solutions = [optp.Individual(x.vector)]

        problem.batch_evaluate(solutions)

        return [solutions[0].objective_values]


class TestCEC2005(unittest.TestCase):
    """Tests from CEC2005 library"""

    def run_test_problem(self, nr_gen, function):
        global func
        func = function
        problem = CEC2005_TEST_Problems()
        algorithm = PSO(problem)
        algorithm.options['max_population_number'] = nr_gen
        algorithm.options['max_population_size'] = 100
        algorithm.run()

        b = Results(problem)
        optimum = b.find_minimum('F_1')  # Takes last cost function
        self.assertAlmostEqual(optimum.costs[0], func.bias, 0)
        del func

    def test_shifted_sphere(self):
        self.run_test_problem(300, optp.cec2005.F1(2))

    def test_shifted_double_sum(self):
        self.run_test_problem(500, optp.cec2005.F2(2))

    #def test_schwefel(self):  # !
    #    self.run_test_problem(400, optp.cec2005.F5(2))

    # the success rate is about 80% reference  Maurice Clerc - A method to improve PSO
    #def test_shifted_rosenbrock(self):
    #    self.run_test_problem(100, optp.cec2005.F6(2))

    # no bounds
    #def test_shifted_griewank(self):
    #    self.run_test_problem(300, optp.cec2005.F7(2))

    def test_shifted_rastrigin(self):
        self.run_test_problem(300, optp.cec2005.F9(2))

    def test_shifted_rot_weierstrass(self):
        self.run_test_problem(300, optp.cec2005.F11(2))

    def test_fletcherpowell(self):  # schwefel
        self.run_test_problem(300, optp.cec2005.F12(2))

    def test_f8f2(self):
        self.run_test_problem(300, optp.cec2005.F13(2))

    def test_expanded_f6(self):
        self.run_test_problem(300, optp.cec2005.F14(2))

#    def test_hybrid_composition(self):
#        self.run_test_problem(300, optp.cec2005.F15(2))

#    def test_rotated_f15(self):
#        self.run_test_problem(300, optp.cec2005.F16(2))

    def test_f17_with_noise(self):
        self.run_test_problem(1000, optp.cec2005.F17(2))

if __name__ == '__main__':
    unittest.main()



import textwrap
import re
import paramiko
import os
import datetime
import time
import ntpath
from string import Template
from xml.dom import minidom
from artap.environment import Enviroment

from abc import ABCMeta, abstractmethod
from .utils import ConfigDictionary
from shutil import copyfile

from logging import NullHandler, getLogger
getLogger('paramiko.transport').addHandler(NullHandler())


class Executor(metaclass=ABCMeta):

    """
    Function is a class representing objective or cost function for
    optimization problems.
    """

    def __init__(self, problem):
        self.problem = problem

        self.options = ConfigDictionary()
        # parse method
        self.parse_results = None

    @abstractmethod
    def eval(self, x):
        # set default parse method from problem (can be overridden)
        if self.parse_results is None and "parse_results" in dir(self.problem):
            self.parse_results = self.problem.parse_results

    @staticmethod
    def _join_parameters_names(parameters, sep=","):
        param_names_string = ""

        for parameter in parameters:
            param_names_string += str(parameter['name']) + sep

        # remove last sep
        if (len(parameters)) > len(sep):
            param_names_string = param_names_string[:-len(sep)]

        return param_names_string

    @staticmethod
    def _join_parameters_values(x, sep=","):
        param_values_string = ""
        for number in x:
            param_values_string += str(number) + sep

        # remove last sep
        if (len(x)) > len(sep):
            param_values_string = param_values_string[:-len(sep)]

        return param_values_string

    @staticmethod
    def _remove_dir(local_dir):
        files = os.listdir(path=local_dir)
        for f in files:
            filepath = os.path.join(local_dir, f)
            os.remove(filepath)

        os.rmdir(local_dir)


class LocalComsolExecutor(Executor):

    comsol_command = Template("comsol batch -inputfile $input_file -nosave -pname $param_names -plist $param_values")

    def __init__(self, problem, problem_file, input_files=None, output_files=None):
        super().__init__(problem)
        self.model_file = ntpath.basename(problem_file)
        copyfile(problem_file, self.problem.working_dir + self.model_file)

        self.output_files = output_files
        self.input_files = input_files

    def eval(self, individual):
        super().eval(individual)

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)

        run_string = self.comsol_command.substitute(input_file=self.model_file,
                                                    param_names=param_names_string,
                                                    param_values=param_values_string)

        # run command
        current_path = os.getcwd()
        os.chdir(self.problem.working_dir)
        os.system(run_string)
        os.chdir(current_path)
        output_files = []
        for file in self.output_files:
            output_files.append(self.problem.working_dir + file)
        result = self.parse_results(output_files, individual)
        return result

    @staticmethod
    def parse_condor_log(content):
        pre_hack = '<?xml version="1.0"?><!DOCTYPE classads SYSTEM "classads.dtd"><classads>'
        post_hack = "</classads>"

        xml_doc = minidom.parseString(pre_hack + content + post_hack)
        event_time = ""
        state = ""
        cluster = 0

        class_ads = xml_doc.getElementsByTagName('c')
        execute_host = ""  # only in "ExecuteEvent"
        for c in class_ads:
            ads = c.getElementsByTagName('a')

            state = ""
            event_time = ""
            cluster = 0
            for a in ads:
                n = a.attributes['n'].value

                if n == "MyType":
                    s = a.getElementsByTagName('s')[0].firstChild.nodeValue

                    if s == "JobTerminatedEvent":
                        state = "Completed"
                    elif s == "SubmitEvent":
                        state = "Submitted"
                    elif s == "ExecuteEvent":
                        state = "Executed"
                    elif s == "JobImageSizeEvent":
                        state = "Running"
                    elif s == "JobDisconnectedEvent":
                        state = "Disconnected"
                    elif s == "JobAbortedEvent":
                        state = "JobAbortedEvent"
                    elif s == "JobHeldEvent":
                        state = "Held"

                if n == "EventTime":
                    s = a.getElementsByTagName('s')[0].firstChild.nodeValue
                    event_time = s.replace("T", " ")

                if n == "ExecuteHost":
                    s = a.getElementsByTagName('s')[0].firstChild.nodeValue
                    execute_host = s.split(":")[0][1:]

                if n == "Cluster":
                    s = a.getElementsByTagName('i')[0].firstChild.nodeValue
                    cluster = int(s)

            # print(event_time + ": " + state + "\t" + execute_host)

        return [event_time, state, cluster, execute_host]


class RemoteExecutor(Executor):
    """
        Allows distributing of calculation of objective functions.
        """

    # ToDo: Make inheritated classes: Matlab, Comsol ...

    def __init__(self, problem, command, files_to_server, files_from_server=None):
        super().__init__(problem)

        if problem.working_dir is None:
            raise Exception('RemoteExecutor: Problem working directory must be set.')

        self.options.declare(name='hostname', default=Enviroment.ssh_host,
                             desc='Hostname')
        self.options.declare(name='username', default=Enviroment.ssh_login,
                             desc='Username')
        self.options.declare(name='port', default=22, lower=0,
                             desc='Port')
        self.options.declare(name='password', default=None,
                             desc='Password')

        # command
        self.command = command

        # files
        self.input_files = files_to_server
        self.output_files = files_from_server

    @abstractmethod
    def eval(self, individual):
        super().eval(individual)

    def _create_client(self):
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(hostname=self.options["hostname"],
                       username=self.options["username"],
                       password=self.options["password"])

        return client

    def _init_remote(self, directory):
        try:
            # create client
            client = self._create_client()
            remote_dir = self._create_dir_on_remote(directory=directory, client=client)
            client.close()
            return remote_dir
        except paramiko.ssh_exception.SSHException:
            self.problem.logger.error("Cannot create remote directory '{}' on ''".format(directory,
                                                                                         self.options["hostname"]))

    @staticmethod
    def _transfer_file_to_remote(source_file, destination_file, remote_dir, client):
        source = source_file
        dest = remote_dir + os.sep + destination_file

        sftp = paramiko.SFTPClient.from_transport(client.get_transport())
        sftp.put(source, dest)

    @staticmethod
    def _transfer_file_from_remote(source_file, destination_file, remote_dir, client):
        dest = destination_file
        source = remote_dir + os.sep + source_file
        sftp = paramiko.SFTPClient.from_transport(client.get_transport())
        sftp.get(source, dest)

    def _read_file_from_remote(self, source_file, remote_dir, client):
        try:
            source = remote_dir + os.sep + source_file
            sftp = paramiko.SFTPClient.from_transport(client.get_transport())
            remote_file = sftp.open(source)
            result = remote_file.read().decode("utf-8")
            return result
        except IOError:
            self.problem.logger.error("Remote file '{}' doesn't exists.".format(source_file))

    @staticmethod
    def _create_dir_on_remote(directory, client):
        d = datetime.datetime.now()
        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")

        # projects directory
        client.exec_command("mkdir " + directory)

        # working directory
        wd = directory + os.sep + "artap-" + ts
        client.exec_command("mkdir " + wd)

        return wd

    @staticmethod
    def _create_file_on_remote(source_file, content, remote_dir, client):
        source = remote_dir + os.sep + source_file
        sftp = paramiko.SFTPClient.from_transport(client.get_transport())
        with sftp.open(source, 'w') as file:
            file.write(content)

    @staticmethod
    def _remove_remote_dir(remote_dir, client):
        sftp = paramiko.SFTPClient.from_transport(client.get_transport())

        files = sftp.listdir(path=remote_dir)
        for f in files:
            filepath = os.path.join(remote_dir, f)
            sftp.remove(filepath)

        sftp.rmdir(remote_dir)

    def _run_command_on_remote(self, command, remote_dir, client, suppress_stdout=True, suppress_stderr=False):
        # Run ssh command
        if remote_dir == "":
            stdin, stdout, stderr = client.exec_command(command)
        else:
            stdin, stdout, stderr = client.exec_command("cd " + remote_dir + "; " + command)

        # output
        output = ""
        for line in stdout:
            output += line.strip('\n')

        if not suppress_stdout:
            lines = ""
            for line in stdout:
                lines += line + '\n'
            self.problem.logger.info(lines)

        if not suppress_stderr:
            lines = ""
            for line in stdout:
                lines += line + '\n'
            if len(lines) > 0:
                self.problem.logger.error(lines)

        return output

    def _transfer_files_to_remote(self, remote_dir, client):
        # transfer input files
        if self.input_files:
            for file in self.input_files:
                self._transfer_file_to_remote(self.problem.working_dir + os.sep + file, "." + os.sep + file,
                                              remote_dir=remote_dir, client=client)

    def _transfer_files_from_remote(self, client):
        pass


class CondorJobExecutor(RemoteExecutor):
    """
    Allows distributing of calculation of objective functions.
    """

    condor_command = """condor_submit remote.job"""

    condor_job_template = Template(textwrap.dedent("""\
        universe = vanilla
        initialdir = ./
        executable = $run_file
        arguments = $arguments
        requirements = (OpSys == "LINUX" && Arch == "X86_64")

        input = $input_file
        output = $log_file
        log = $$(cluster).condor_log
        error = $$(cluster)_$$(process).condor_err

        request_cpus = 10
        request_memory = 10 GB
        log_xml = true
        should_transfer_files = yes
        when_to_transfer_output = ON_EXIT
        transfer_input_files = $input_files
        transfer_output_files = $output_files
        transfer_executable = true
        queue"""))

    def __init__(self, problem, files_to_condor, files_from_condor=None):
        super().__init__(problem, self.condor_command, files_to_condor, files_from_condor)

        # set default host
        self.options["hostname"] = Enviroment.condor_host
        self.job_file = None
        self.condor_executable = None

    @abstractmethod
    def _create_job_file(self, remote_dir, individual, client):
        pass

    def eval(self, individual):
        super().eval(individual)

        # init remote
        remote_dir = self._init_remote(directory="htcondor")

        success = False
        while not success:
            try:
                # create client
                client = self._create_client()

                # apply template for remote.job
                self._create_job_file(remote_dir, individual, client)

                # Create
                self._create_file_on_remote("remote.job", self.job_file, remote_dir=remote_dir, client=client)
                self._create_file_on_remote("run.sh", self.condor_executable, remote_dir=remote_dir,
                                            client=client)

                # transfer supplementary files, input and model file
                self._transfer_files_to_remote(remote_dir, client)

                # submit job
                output = self._run_command_on_remote("condor_submit remote.job", remote_dir=remote_dir, client=client)

                # process id
                process_id = re.search(r'cluster \d+', output).group().split(" ")[1]

                event = ""
                start = time.time()
                while (event != "Completed") and (event != "Held"):
                    content = self._read_file_from_remote("{}.condor_log".format(process_id),
                                                          remote_dir=remote_dir, client=client)
                    state = LocalComsolExecutor.parse_condor_log(content)

                    if state[1] != event:
                        self.problem.logger.info("Job {} is '{}' at {}".format(state[2], state[1], state[3]))
                        event = state[1]

                    if event == "Completed":
                        # stop while cycle
                        break

                    if (event == "Held") or (event == "JobAbortedEvent"):
                        self.problem.logger.error("Job {} is '{}' at {}".format(state[2], state[1], state[3]))
                        # read log
                        content_log = self._read_file_from_remote("{}.log".format(self.output_files[0]),
                                                                  remote_dir=remote_dir, client=client)
                        self.problem.logger.error(content_log)
                        # remove job
                        self._run_command_on_remote("condor_rm {}".format(process_id),
                                                    remote_dir=remote_dir, client=client)
                        # TODO: abort computation - no success?
                        raise RuntimeError

                    end = time.time()
                    if (end - start) > self.problem.options["time_out"]:
                        raise TimeoutError

                if event == "Completed":
                    output_files = []
                    d = datetime.datetime.now()
                    ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")
                    path = self.problem.working_dir + 'artap' + ts
                    os.system("mkdir {0}".format(path))

                    for file in self.output_files:
                        self._transfer_file_from_remote(source_file=file, destination_file=path + os.sep + file,
                                                        remote_dir=remote_dir, client=client)
                        output_files.append(path + os.sep + file)
                    success = True
                    result = self.parse_results(output_files, individual)
                    self._remove_remote_dir(remote_dir, client=client)

                    if self.problem.options['save_data_files'] is False:
                        self._remove_dir(path)
                else:
                    assert 0

                # self._remove_remote_dir(remote_dir=remote_dir, client=client)

                client.close()

                return result

            except ConnectionError as e:
                print(e)
                time.sleep(1.0)
                continue


class CondorComsolJobExecutor(CondorJobExecutor):
    """

    """
    arguments = Template("-inputfile $input_file -nosave -pname $param_names -plist $param_values")
    executable = textwrap.dedent("""\
        #!/bin/sh
        /opt/comsol-5.4/bin/comsol batch $@
        """)

    def __init__(self, problem, model_file, files_from_condor=None):
        self.model_file = ntpath.basename(model_file)
        super().__init__(problem, [self.model_file], files_from_condor)
        copyfile(model_file, self.problem.working_dir + self.model_file)

    def _create_job_file(self, remote_dir, individual, client):
        output_files_str = ",".join(self.output_files)
        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)
        arguments = self.arguments.substitute(input_file=os.path.basename(self.model_file),
                                              param_names=param_names_string,
                                              param_values=param_values_string)

        self.job_file = self.condor_job_template.substitute(input_file=os.path.basename(self.model_file),
                                                            run_file="run.sh",
                                                            arguments=arguments,
                                                            input_files=self.model_file,
                                                            output_files=output_files_str,
                                                            log_file="{}.log".format(self.output_files[0]))
        self.condor_executable = self.executable


class CondorMatlabJobExecutor(CondorJobExecutor):

    executable = textwrap.dedent("""\
    #!/bin/sh
    # args
    s=$@

    /opt/matlab-R2018b/bin/matlab -nodisplay -nosplash -nodesktop -r ${s%.m}""")

    def __init__(self, problem, script, parameter_file, files_from_condor=None):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], files_from_condor)
        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = ",".join(self.output_files)
        condor_input_files = self.parameter_file + ", " + self.input_files[0]
        # create input file with parameters

        if self.parameter_file:
            # parameters
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            # create remote file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir, client=client)

        self.job_file = self.condor_job_template.substitute(input_file=os.path.basename(self.script),
                                                            run_file="run.sh",
                                                            arguments=self.script,
                                                            input_files=condor_input_files,
                                                            output_files=condor_output_files,
                                                            log_file="{}.log".format(self.output_files[0]))
        self.condor_executable = self.executable


class CondorPythonJobExecutor(CondorJobExecutor):

    executable = textwrap.dedent("""\
        #!/bin/sh
        # args
                      
        python3 $@""")

    def __init__(self, problem, script, parameter_file, output_files=None):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], output_files)
        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = ",".join(self.output_files)
        condor_input_files = self.script
        # create input file with parameters

        if self.parameter_file:
            # parameters
            # create remote file
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            condor_input_files += ", " + self.parameter_file
            self.arguments = self.script + " " + self.parameter_file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir,
                                        client=client)
        else:
            param_values_string = Executor._join_parameters_values(individual.vector, ",")
            self.arguments = self.script + " " + param_values_string

        self.job_file = self.condor_job_template.substitute(input_file=os.path.basename(self.script),
                                                            run_file="run.sh",
                                                            arguments=self.arguments,
                                                            input_files=condor_input_files,
                                                            output_files=condor_output_files,
                                                            log_file="{}.log".format(self.output_files[0]))
        self.condor_executable = self.executable



from .problem import Problem
from .utils import ConfigDictionary
from .job import JobSimple
from .population import Population
from .individual import Individual

from abc import ABCMeta, abstractmethod
import numpy as np

from joblib import Parallel, delayed


class Algorithm(metaclass=ABCMeta):
    """ Base class for optimization algorithms. """

    def __init__(self, problem: Problem, name="Algorithm"):
        self.name = name
        self.problem = problem
        self.parameters = problem.parameters
        self.options = ConfigDictionary()

        self.options.declare(name='verbose_level', default=1, lower=0,
                             desc='Verbose level')
        self.options.declare(name='calculate_gradients', default=False,
                             desc='Enable calculating of gradients')
        # max(int(2 / 3 * cpu_count(), 1)
        self.options.declare(name='max_processes', default=1,
                             desc='Max running processes')

        # initial population size
        self.population_size = 0

    @abstractmethod
    def run(self):
        pass

    def gen_initial_population(self, is_archive=False):
        individuals = self.generator.generate()
        # set current size
        self.population_size = len(individuals)
        # evaluate individuals
        self.evaluate(individuals)

        if is_archive:
            population = Population(individuals, individuals)
        else:
            population = Population(individuals)
        return population

    def evaluate(self, individuals: list):
        if self.options["max_processes"] > 1:
            self.evaluate_parallel(individuals)
        else:
            self.evaluate_serial(individuals)
        n_failed = 0
        for individual in individuals:
            if individual.state == Individual.State.FAILED:
                n_failed += 1
                individuals.remove(individual)

    def evaluate_serial(self, individuals: list):
        job = JobSimple(self.problem)
        for individual in individuals:
            if individual.state == Individual.State.EMPTY:
                job.evaluate(individual)

    def evaluate_parallel(self, individuals: list):
        # simple parallel loop
        job = JobSimple(self.problem)
        Parallel(n_jobs=self.options["max_processes"], verbose=1, require='sharedmem')(delayed(job.evaluate)(individual)
                                                                                       for individual in individuals)



        """
        manager = Manager()
        processes = []
        i = 0
        n = len(individuals)
        while i < n:
            j = 0
            shared_list = manager.list([])
            queue = Queue()
            job = JobQueue(self.problem, shared_list, queue)
            while (j < self.options["max_processes"]) and ((i + j) < n):
                individual = individuals[i + j]
                if not individual.is_evaluated:
                    p = Process(target=job.evaluate, args=(individual,))
                    processes.append(p)
                    p.start()
                    shared_list.append([p.pid, individual])
                # increment index
                j += 1

            for process in processes:
                process.join()

            processes = []
            i += j

            for item in range(queue.qsize()):
                local_individual = queue.get()
                # TODO: rewrite !!! - very slow
                for individual in individuals:
                    if local_individual.vector == individual.vector:
                        # update individual
                        individual.sync(local_individual)

                        # write to datastore
                        if self.problem.options["save_level"] == "individual":
                            self.problem.data_store.write_individual(individual)
            queue.close()
            queue.join_thread()
        """

    def evaluate_gradient(self, individuals: list, gradient_individuals: list):
        # Todo: Make operator Gradient consisting of current gradient generator and evaluate_gradient
        # individuals.sort(key=lambda x: abs(x.depends_on))
        n_params = len(individuals[0].vector)
        for j in range(len(individuals)):
            table = []
            for i in range(len(gradient_individuals)):
                individual = gradient_individuals[i]
                if individual.depends_on == j:
                    table.append(individual)

            gradient = np.zeros(n_params)
            for k in range(len(table)):
                index = table[k].modified_param
                gradient[index] = ((individuals[j].costs[0] - table[k].costs[0]) / 1e-6)

            individuals[j].gradient = gradient


class DummyAlgorithm(Algorithm):
    """
    Dummy class for testing
    """

    def __init__(self, problem, name='DummyAlgorithm'):
        super().__init__(problem, name)

    def run(self):
        pass



from random import randint
from .problem import Problem
from .population import Population
from .algorithm_genetic import GeneticAlgorithm, GeneticIndividual
from .operators import SwarmMutation, DummySelection, RandomGeneration

import time


class PSO(GeneticAlgorithm):

    def __init__(self, problem: Problem, name="Particle Swarm Algorithm"):
        super().__init__(problem, name)
        self.n = self.options['max_population_size']
        self.mutator = SwarmMutation(self.problem.parameters)
        self.selector = DummySelection(self.problem.parameters, self.problem.signs)

    def run(self):
        # set random generator
        self.generator = RandomGeneration(self.problem, individual_class=GeneticIndividual)
        self.generator.init(self.options['max_population_size'])

        population = self.gen_initial_population()

        for individual in population.individuals:
            self.mutator.evaluate_pso(individual)

        self.selector.sorting(population.individuals)
        self.problem.data_store.write_population(population)

        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))

        i = 0
        while i < self.options['max_population_number']:
            offsprings = self.selector.select(population.individuals)

            pareto_front = []
            for individual in offsprings:
                if individual.front_number == 1:
                    pareto_front.append(individual)

            for individual in offsprings:
                index = randint(0, len(pareto_front) - 1)  # takes random individual from Pareto front
                best_individual = pareto_front[index]
                if best_individual is not individual:
                    self.mutator.update(best_individual)
                    self.mutator.mutate(individual)

            self.evaluate(offsprings)

            for individual in offsprings:
                self.mutator.evaluate_pso(individual)

            self.selector.sorting(offsprings)
            population = Population(offsprings)
            self.problem.data_store.write_population(population)

            i += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))
